"""The pyramid match kernel as in :cite:`nikolentzos2017matching`."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
import collections
import warnings

import numpy as np

from itertools import chain

from scipy.sparse import csr_matrix
from scipy.sparse.linalg import eigs

from grakel.graph import Graph
from grakel.kernels import Kernel

# Python 2/3 cross-compatibility import
from six import itervalues
from six import iteritems


class PyramidMatch(Kernel):
    """Pyramid match kernel class.

    Kernel defined in :cite:`nikolentzos2017matching`

    Parameters
    ----------
    with_labels : bool, default=True
        A flag that determines if the kernel computation will consider labels.

    L : int, default=4
        Pyramid histogram level.

    d : int, default=6
        The dimension of the hypercube.

    Attributes
    ----------
    _num_labels : int
        The number of distinct labels, on the fit data.

    _labels : dict
        A dictionary of label enumeration, made from fitted data.

    """

    _graph_format = "adjacency"

    def __init__(self, n_jobs=None,
                 normalize=False,
                 verbose=False,
                 with_labels=True,
                 L=4,
                 d=6):
        """Initialise a `pyramid_match` kernel."""
        super(PyramidMatch, self).__init__(n_jobs=n_jobs,
                                           normalize=normalize,
                                           verbose=verbose)

        self.with_labels = with_labels
        self.L = L
        self.d = d
        self._initialized.update({"d": False, "L": False, "with_labels": False})

    def initialize(self):
        """Initialize all transformer arguments, needing initialization."""
        super(PyramidMatch, self).initialize()

        if not self._initialized["with_labels"]:
            if type(self.with_labels) != bool:
                raise TypeError('with labels must be a boolean variable')
            self._initialized["with_labels"] = True

        if not self._initialized["L"]:
            if type(self.L) is not int or self.L < 0:
                raise TypeError('L: the number of levels must be an integer '
                                'bigger equal to 0')
            self._initialized["L"] = True

        if not self._initialized["d"]:
            if type(self.d) is not int or self.d < 1:
                raise TypeError('d: hypercube dimension must be an '
                                'integer bigger than 1')
            self._initialized["d"] = True

    def parse_input(self, X):
        """Parse and create features for pyramid_match kernel.

        Parameters
        ----------
        X : iterable
            For the input to pass the test, we must have:
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that correspond to the given
            graph format). A valid input also consists of graph type objects.

        Returns
        -------
        H : list
            A list of lists of Histograms for all levels for each graph.

        """
        if not isinstance(X, collections.Iterable):
            raise TypeError('input must be an iterable\n')
        else:
            i = 0
            Us = []
            if self.with_labels:
                Ls = []
            for (idx, x) in enumerate(iter(X)):
                is_iter = isinstance(x, collections.Iterable)
                if is_iter:
                    x = list(x)
                if is_iter and (len(x) == 0 or (len(x) >= 1 and not self.with_labels) or
                                (len(x) >= 2 and self.with_labels)):
                    if len(x) == 0:
                        warnings.warn('Ignoring empty element on index: ' + str(idx))
                        continue
                    elif not self.with_labels:
                        x = Graph(x[0], {}, {}, self._graph_format)
                    else:
                        x = Graph(x[0], x[1], {}, self._graph_format)
                elif not type(x) is Graph:
                    raise TypeError('each element of X must be either a graph object or a list with '
                                    'at least a graph like object and node labels dict \n')
                A = x.get_adjacency_matrix()
                if self.with_labels:
                    L = x.get_labels(purpose="adjacency")
                i += 1
                if A.shape[0] == 0:
                    Us.append(np.zeros((1, self.d)))
                else:
                    # Perform eigenvalue decomposition.
                    # Rows of matrix U correspond to vertex representations
                    # Embed vertices into the d-dimensional space
                    if A.shape[0] > self.d+1:
                        # If size of graph smaller than d, pad with zeros
                        Lambda, U = eigs(csr_matrix(A, dtype=np.float),
                                         k=self.d, ncv=10*self.d)
                        idx = Lambda.argsort()[::-1]
                        U = U[:, idx]
                    else:
                        Lambda, U = np.linalg.eig(A)
                        idx = Lambda.argsort()[::-1]
                        U = U[:, idx]
                        U = U[:, :self.d]
                    # Replace all components by their absolute values
                    U = np.absolute(U)
                    Us.append((A.shape[0], U))
                if self.with_labels:
                    Ls.append(L)

        if i == 0:
            raise ValueError('parsed input is empty')

        if self.with_labels:
            # Map labels to values between 0 and |L|-1
            # where |L| is the number of distinct labels
            if self._method_calling in [1, 2]:
                self._num_labels = 0
                self._labels = set()
                for L in Ls:
                    self._labels |= set(itervalues(L))
                self._num_labels = len(self._labels)
                self._labels = {l: i for (i, l) in enumerate(self._labels)}
                return self._histogram_calculation(Us, Ls, self._labels)

            elif self._method_calling == 3:
                labels = set()
                for L in Ls:
                    labels |= set(itervalues(L))
                rest_labels = labels - set(self._labels.keys())
                nouveau_labels = dict(chain(iteritems(self._labels),
                                      ((j, i) for (i, j) in enumerate(rest_labels, len(self._labels)))))
                return self._histogram_calculation(Us, Ls, nouveau_labels)
        else:
            return self._histogram_calculation(Us)

    def _histogram_calculation(self, Us, *args):
        """Calculate histograms.

        Parameters
        ----------
        Us : list
            List of tuples with the first element corresponding to the
            number of vertices of a graph and the second to it's
            corresponding to vertex embeddings on the d-dimensional space.

        Ls : list, optional
            List of labels corresponding to each graph.
            If provided the histograms are calculated with labels.

        Labels : dict, optional
            A big dictionary with enumeration of labels.

        Returns
        -------
        Hs : list
            List of histograms for each graph.

        """
        Hs = list()
        if len(args) == 0:
            for (i, (n, u)) in enumerate(Us):
                du = list()
                if n > 0:
                    for j in range(self.L):
                        # Number of cells along each dimension at level j
                        k = 2**j
                        # Determines the cells in which each vertex lies
                        # along each dimension since nodes lie in the unit
                        # hypercube in R^d
                        D = np.zeros((self.d, k))
                        T = np.floor(u*k)
                        T[np.where(T == k)] = k-1
                        for p in range(u.shape[0]):
                            if p >= n:
                                break
                            for q in range(u.shape[1]):
                                # Identify the cell into which the i-th
                                # vertex lies and increase its value by 1
                                D[q, int(T[p, q])] += 1
                        du.append(D)
                Hs.append(du)

        elif len(args) > 0:
            Ls = args[0]
            Labels = args[1]
            num_labels = len(Labels)
            for (i, ((n, u), L)) in enumerate(zip(Us, Ls)):
                du = list()
                if n > 0:
                    for j in range(self.L):
                        # Number of cells along each dimension at level j
                        k = 2**j
                        # To store the number of vertices that are assigned
                        # a specific label and lie in each of the 2^j cells
                        # of each dimension at level j
                        D = np.zeros((self.d*num_labels, k))
                        T = np.floor(u*k)
                        T[np.where(T == k)] = k-1
                        for p in range(u.shape[0]):
                            if p >= n:
                                break
                            for q in range(u.shape[1]):
                                # Identify the cell into which the i-th
                                # vertex lies and increase its value by 1
                                D[Labels[L[p]]*self.d + q, int(T[p, q])] += 1
                        du.append(D)
                Hs.append(du)
        return Hs

    def pairwise_operation(self, x, y):
        """Calculate a pairwise kernel between two elements.

        Parameters
        ----------
        x, y : dict
            Histograms as produced by `parse_input`.

        Returns
        -------
        kernel : number
            The kernel value.

        """
        k = 0
        if len(x) != 0 and len(y) != 0:
            intersec = np.zeros(self.L)
            for (p, xp, yp) in zip(range(self.L), x, y):
                # Calculate histogram intersection
                # (eq. 6 in :cite:`nikolentzos2017matching`)
                if xp.shape[0] < yp.shape[0]:
                    xpp, ypp = xp, yp[:xp.shape[0], :]
                elif yp.shape[0] < xp.shape[0]:
                    xpp, ypp = xp[:yp.shape[0], :], yp
                else:
                    xpp, ypp = xp, yp
                intersec[p] = np.sum(np.minimum(xpp, ypp))
                k += intersec[self.L-1]
                for p in range(self.L-1):
                    # Computes the new matches that occur at level p.
                    # These matches weight less than those that occur at
                    # higher levels (e.g. p+1 level)
                    k += (1.0/(2**(self.L-p-1)))*(intersec[p]-intersec[p+1])
        return k



"""Init file for the _isomorphism submodule project."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# This file is a modification and extension of the [GNU LPGL] licensed
# PyBliss which can be found at: http://www.tcs.hut.fi/Software/bliss/
# PyBliss and Bliss are copyright of their respective owners.
# License: BSD 3 clause"
from grakel.kernels._isomorphism.bliss import Graph

__all__ = [
    "Graph",
]



"""Multiscale Laplacian Graph Kernel as defined in :cite:`kondor2016multiscale`."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
# Python 2/3 cross-compatibility import
from __future__ import print_function

import collections
import warnings
import numpy as np

from numbers import Real
from math import exp

from sklearn.utils import check_random_state

from numpy.linalg import eig
from numpy.linalg import inv
from numpy.linalg import multi_dot
from numpy.linalg import eigvals

from grakel.graph import Graph
from scipy.sparse.csgraph import laplacian

from grakel.kernels import Kernel

positive_eigenvalue_limit = float("+1e-6")


class MultiscaleLaplacian(Kernel):
    """Laplacian Graph Kernel as proposed in :cite:`kondor2016multiscale`.

    Parameters
    ----------
    random_state :  RandomState or int, default=None
        A random number generator instance or an int to initialize a RandomState as a seed.

    L : int, default=3
        The number of neighborhoods.

    gamma : Real, default=0.01
        A smoothing parameter of float value.

    heta : float, default=0.01
        A smoothing parameter of float value.

    P : int, default=10
        Restrict the maximum number of eigenvalues, taken on eigenvalue decomposition.

    n_samples : int, default=50
        The number of vertex samples.

    Attributes
    ----------
    random_state_ : RandomState
        A RandomState object handling all randomness of the class.

    _data_level : dict
        A dictionary containing the feature basis information needed
        for each level calculation on transform.

    """

    _graph_format = "adjacency"

    def __init__(self,
                 n_jobs=None,
                 normalize=False, verbose=False,
                 random_state=None,
                 L=3,
                 P=10,
                 gamma=0.01,
                 heta=0.01,
                 n_samples=50):
        """Initialise a `multiscale_laplacian` kernel."""
        super(MultiscaleLaplacian, self).__init__(
            n_jobs=n_jobs,
            normalize=normalize,
            verbose=verbose)

        self.random_state = random_state
        self.gamma = gamma
        self.heta = heta
        self.L = L
        self.P = P
        self.n_samples = n_samples
        self._initialized.update({"random_state": False, "gamma": False,
                                  "heta": False, "L": False, "n_samples": False, "P": False})

    def initialize(self):
        """Initialize all transformer arguments, needing initialization."""
        super(MultiscaleLaplacian, self).initialize()

        if not self._initialized["random_state"]:
            self.random_state_ = check_random_state(self.random_state)
            self._initialized["random_state"] = True

        if not self._initialized["gamma"]:
            if not isinstance(self.gamma, Real):
                raise TypeError('gamma must be a real number')
            elif self.gamma == .0:
                warnings.warn('with zero gamma the calculation may crash')
            elif self.gamma < 0:
                raise TypeError('gamma must be positive')
            self._initialized["gamma"] = True

        if not self._initialized["heta"]:
            if not isinstance(self.heta, Real):
                raise TypeError('heta must be a real number')
            elif self.heta == .0:
                warnings.warn('with zero heta the calculation may crash')
            elif self.heta < 0:
                raise TypeError('heta must be positive')
            self._initialized["heta"] = True

        if not self._initialized["L"]:
            if type(self.L) is not int:
                raise TypeError('L must be an integer')
            elif self.L < 0:
                raise TypeError('L must be positive')
            self._initialized["L"] = True

        if not self._initialized["n_samples"]:
            if type(self.n_samples) is not int or self.n_samples <= 0:
                raise TypeError('n_samples must be a positive integer')
            self._initialized["n_samples"] = True

        if not self._initialized["P"]:
            if type(self.P) is not int or self.P <= 0:
                raise TypeError('P must be a positive integer')
            self._initialized["P"] = True

    def parse_input(self, X):
        """Fast ML Graph Kernel.

        See supplementary material :cite:`kondor2016multiscale`, algorithm 1.

        Parameters
        ----------
        X : iterable
            For the input to pass the test, we must have:
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that correspond to the given
            graph format). A valid input also consists of graph type objects.

        Returns
        -------
        out : list
            A list of tuples with S matrices inverses
            and their 4th-root determinants.

        """
        if not isinstance(X, collections.Iterable):
            raise TypeError('input must be an iterable\n')
        else:
            ng = 0
            out = list()
            data = dict()
            neighborhoods = dict()
            for (idx, x) in enumerate(iter(X)):
                is_iter = False
                if isinstance(x, collections.Iterable):
                    is_iter, x = True, list(x)
                if is_iter and len(x) in [0, 2, 3]:
                    if len(x) == 0:
                        warnings.warn('Ignoring empty element ' +
                                      'on index: '+str(idx))
                        continue
                    else:
                        x = Graph(x[0], x[1], {}, self._graph_format)
                elif type(x) is Graph:
                    x.desired_format(self._graph_format)
                else:
                    raise TypeError('each element of X must be either a '
                                    'graph or an iterable with at least 1 '
                                    'and at most 3 elements\n')
                phi_d = x.get_labels()
                A = x.get_adjacency_matrix()
                try:
                    phi = np.array([list(phi_d[i]) for i in range(A.shape[0])])
                except TypeError:
                    raise TypeError('Features must be iterable and castable '
                                    'in total to a numpy array.')

                Lap = laplacian(A).astype(float)
                _increment_diagonal_(Lap, self.heta)
                data[ng] = {0: A, 1: phi, 2: inv(Lap)}
                neighborhoods[ng] = x
                ng += 1

            if ng == 0:
                raise ValueError('parsed input is empty')

            # Define a function for calculating the S's of subgraphs of each iteration
            def calculate_C(k, j, l):
                if type(neighborhoods[k]) is Graph:
                    neighborhoods[k] = neighborhoods[k].produce_neighborhoods(
                        r=self.L, sort_neighbors=False)

                indexes = neighborhoods[k][l][j]
                L = laplacian(data[k][0][indexes, :][:, indexes]).astype(float)
                _increment_diagonal_(L, self.heta)
                U = data[k][1][indexes, :]
                S = multi_dot((U.T, inv(L), U))
                _increment_diagonal_(S, self.gamma)

                return (inv(S), np.sum(np.log(np.real(eigvals(S)))))

            if self._method_calling == 1:
                V = [(k, j) for k in range(ng)
                     for j in range(data[k][0].shape[0])]

                ns = min(len(V), self.n_samples)

                self.random_state_.shuffle(V)
                vs = V[:ns]
                phi_k = np.array([data[k][1][j, :] for (k, j) in vs])

                # w the eigen vectors, v the eigenvalues
                K = phi_k.dot(phi_k.T)

                # Calculate eigenvalues
                v, w = eig(K)
                v, w = np.real(v), np.real(w.T)

                # keep only the positive
                vpos = np.argpartition(v, -self.P)[-self.P:]
                vpos = vpos[np.where(v[vpos] > positive_eigenvalue_limit)]

                # ksi.shape = (k, Ns) * (Ns, P)
                ksi = w[vpos].dot(phi_k).T / np.sqrt(v[vpos])
                for j in range(ng):
                    # (n_samples, k) * (k, P)
                    data[j][1] = data[j][1].dot(ksi)
                self._data_level = {0: ksi}
                for l in range(1, self.L+1):
                    # Take random samples from all the vertices of all graphs
                    self.random_state_.shuffle(V)
                    vs = V[:ns]

                    # Compute the reference subsampled Gram matrix
                    K_proj = {k: np.zeros(shape=(data[k][0].shape[0], ns)) for k in range(ng)}
                    K, C = np.zeros(shape=(len(vs), len(vs))), dict()
                    for (m, (k, j)) in enumerate(vs):
                        C[m] = calculate_C(k, j, l)
                        K_proj[k][j, m] = K[m, m] = self.pairwise_operation(C[m], C[m])
                        for (s, (k2, j2)) in enumerate(vs):
                            if s < m:
                                K[s, m] = K[m, s] \
                                        = K_proj[k2][j2, m] \
                                        = K_proj[k][j, s] \
                                        = self.pairwise_operation(C[s], C[m])
                            else:
                                break

                    # Compute the kernels of the relations of the reference to everything else
                    for (k, j) in V[ns:]:
                        for (m, _) in enumerate(vs):
                            K_proj[k][j, m] = self.pairwise_operation(C[m], calculate_C(k, j, l))

                    # w the eigen vectors, v the eigenvalues
                    v, w = eig(K)
                    v, w = np.real(v), np.real(w.T)

                    # keep only the positive
                    vpos = np.argpartition(v, -self.P)[-self.P:]
                    vpos = vpos[np.where(v[vpos] > positive_eigenvalue_limit)]

                    # Q shape=(k, P)
                    Q = w[vpos].T / np.sqrt(v[vpos])
                    for j in range(ng):
                        # (n, ns) * (ns, P)
                        data[j][1] = K_proj[j].dot(Q)
                    self._data_level[l] = (C, Q)

            elif self._method_calling == 3:
                ksi = self._data_level[0]
                for j in range(ng):
                    # (n, k) * (k, P)
                    data[j][1] = data[j][1].dot(ksi)

                for l in range(1, self.L+1):
                    C, Q = self._data_level[l]
                    for j in range(ng):
                        K_proj = np.zeros(shape=(data[j][0].shape[0], len(C)))
                        for n in range(data[j][0].shape[0]):
                            for m in range(len(C)):
                                K_proj[n, m] = self.pairwise_operation(C[m], calculate_C(j, n, l))
                        data[j][1] = K_proj.dot(Q)

            # Apply the final calculation of S.
            for k in range(ng):
                S = multi_dot((data[k][1].T, data[k][2], data[k][1]))
                _increment_diagonal_(S, self.gamma)
                out.append((inv(S), np.sum(np.log(np.real(eigvals(S))))))

            return out

    def pairwise_operation(self, x, y):
        """FLG calculation for the fast multiscale laplacian.

        Parameters
        ----------
        x, y : tuple
            An np.array of inverse and the log determinant of S
            (for the calculation of S matrices see the algorithm 1
             of the supplement material in cite:`kondor2016multiscale`).

        Returns
        -------
        kernel : number
            The FLG core kernel value.

        """
        S_inv_x, log_det_x = x
        S_inv_y, log_det_y = y

        # Calculate the result in term of logs
        log_detS = -np.sum(np.log(np.real(eigvals(S_inv_x + S_inv_y))))
        logr = (log_detS - 0.5*(log_det_x + log_det_y))/2.0

        if logr < -30:
            return .0
        else:
            return exp(logr)


def _increment_diagonal_(A, value):
    """Increment the diagonal of an array by a value.

    Parameters
    ----------
    A : np.array
        The array whose diagonal will be extracted.

    value : number
        The value that will be incremented on the diagonal.


    Returns
    -------
    None.

    """
    d = A.diagonal()
    d.setflags(write=True)
    d += value



"""The base file for loading default datasets."""
# Python 2/3 cross-compatibility import
from __future__ import print_function

import os
import shutil
import zipfile
import ssl
try:
    # Python 2
    from urllib2 import HTTPError
    from urllib2 import urlopen
except ImportError:
    # Python 3+
    from urllib.error import HTTPError
    from urllib.request import urlopen

import numpy as np

from shutil import copyfileobj

from collections import Counter

from sklearn.utils import Bunch

from grakel.graph import Graph

global datasets_metadata, symmetric_dataset

dataset_metadata = {
    "AIDS": {"nl": True, "el": True, "na": True, "ea": False,
             "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
             "morris/graphkerneldatasets/AIDS.zip"},
    "BZR": {"nl": True, "el": False, "na": True, "ea": False,
            "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
            "morris/graphkerneldatasets/BZR.zip"},
    "BZR_MD": {"nl": True, "el": True, "na": False, "ea": True,
               "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
               "morris/graphkerneldatasets/BZR_MD.zip"},
    "COIL-DEL": {"nl": False, "el": True, "na": True, "ea": False,
                 "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/" +
                 "graphkerneldatasets/COIL-DEL.zip"},
    "COIL-RAG": {"nl": False, "el": False, "na": True, "ea": True,
                 "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/" +
                 "graphkerneldatasets/COIL-RAG.zip"},
    "COLLAB": {"nl": False, "el": False, "na": False, "ea": False,
               "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/" +
               "graphkerneldatasets/COLLAB.zip"},
    "COX2": {"nl": True, "el": False, "na": True, "ea": False,
             "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/" +
             "graphkerneldatasets/COX2.zip"},
    "COX2_MD": {"nl": True, "el": True, "na": False, "ea": True,
                "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/" +
                "graphkerneldatasets/COX2_MD.zip"},
    "DHFR": {"nl": True, "el": False, "na": True, "ea": False,
             "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/" +
             "graphkerneldatasets/DHFR.zip"},
    "DHFR_MD": {"nl": True, "el": True, "na": False, "ea": True,
                "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/" +
                "graphkerneldatasets/DHFR_MD.zip"},
    "ER_MD": {"nl": True, "el": True, "na": False, "ea": True,
              "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/" +
              "graphkerneldatasets/ER_MD.zip"},
    "DD": {"nl": True, "el": False, "na": False, "ea": False,
           "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/" +
           "graphkerneldatasets/DD.zip"},
    "ENZYMES": {"nl": True, "el": False, "na": True, "ea": False,
                "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/" +
                "graphkerneldatasets/ENZYMES.zip"},
    "Cuneiform": {"nl": True, "el": True, "na": True, "ea": True,
                  "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/"
                          "graphkerneldatasets/Cuneiform.zip"},
    "FINGERPRINT": {"nl": False, "el": False, "na": True, "ea": True,
                    "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                    "morris/graphkerneldatasets/Fingerprint.zip"},
    "FIRSTMM_DB": {"nl": True, "el": False, "na": True, "ea": True,
                   "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                   "morris/graphkerneldatasets/FIRSTMM_DB.zip"},
    "FRANKENSTEIN": {"nl": False, "el": False, "na": True, "ea": False,
                     "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
                     "ris/graphkerneldatasets/FRANKENSTEIN.zip"},
    "IMDB-BINARY": {"nl": False, "el": False, "na": False, "ea": False,
                    "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
                    "ris/graphkerneldatasets/IMDB-BINARY.zip"},
    "IMDB-MULTI": {"nl": False, "el": False, "na": False, "ea": False,
                   "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                   "morris/graphkerneldatasets/IMDB-MULTI.zip"},
    "Letter-high": {"nl": False, "el": False, "na": True, "ea": False,
                    "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
                    "ris/graphkerneldatasets/Letter-high.zip"},
    "Letter-low": {"nl": False, "el": False, "na": True, "ea": False,
                   "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
                   "ris/graphkerneldatasets/Letter-low.zip"},
    "Letter-med": {"nl": False, "el": False, "na": True, "ea": False,
                   "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
                   "ris/graphkerneldatasets/Letter-med.zip"},
    "Mutagenicity": {"nl": True, "el": True, "na": False, "ea": False,
                     "link": "https://ls11-www.cs.uni-dortmund.de/peo" +
                     "ple/morris/graphkerneldatasets/Mutagenicity.zip"},
    "MSRC_9": {"nl": True, "el": False, "na": False, "ea": False,
               "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/" +
               "graphkerneldatasets/MSRC_9.zip"},
    "MSRC_21": {"nl": True, "el": False, "na": False, "ea": False,
                "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
                "ris/graphkerneldatasets/MSRC_21.zip"},
    "MSRC_21C": {"nl": True, "el": False, "na": False, "ea": False,
                 "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
                 "ris/graphkerneldatasets/MSRC_21C.zip"},
    "MUTAG": {"nl": True, "el": True, "na": False, "ea": False,
              "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
              "ris/graphkerneldatasets/MUTAG.zip"},
    "NCI1": {"nl": True, "el": False, "na": False, "ea": False,
             "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
             "ris/graphkerneldatasets/NCI1.zip"},
    "NCI109": {"nl": True, "el": False, "na": False, "ea": False,
               "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
               "ris/graphkerneldatasets/NCI109.zip"},
    "PTC_FM": {"nl": True, "el": True, "na": False, "ea": False,
               "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
               "ris/graphkerneldatasets/PTC_FM.zip"},
    "PTC_FR": {"nl": True, "el": True, "na": False, "ea": False,
               "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
               "ris/graphkerneldatasets/PTC_FR.zip"},
    "PTC_MM": {"nl": True, "el": True, "na": False, "ea": False,
               "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
               "ris/graphkerneldatasets/PTC_MM.zip"},
    "PTC_MR": {"nl": True, "el": True, "na": False, "ea": False,
               "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
               "ris/graphkerneldatasets/PTC_MR.zip"},
    "PROTEINS": {"nl": True, "el": False, "na": True, "ea": False,
                 "link": "https://ls11-www.cs.uni-dortmund.de/people/mor" +
                 "ris/graphkerneldatasets/PROTEINS.zip"},
    "PROTEINS_full": {"nl": True, "el": False, "na": True, "ea": False,
                      "link": "https://ls11-www.cs.uni-dortmund.de/people" +
                      "/morris/graphkerneldatasets/PROTEINS_full.zip"},
    "REDDIT-BINARY": {"nl": False, "el": False, "na": False, "ea": False,
                      "link": "https://ls11-www.cs.uni-dortmund.de/people" +
                      "/morris/graphkerneldatasets/REDDIT-BINARY.zip"},
    "REDDIT-MULTI-5K": {"nl": False, "el": False, "na": False, "ea": False,
                        "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                        "morris/graphkerneldatasets/REDDIT-MULTI-5K.zip"},
    "REDDIT-MULTI-12K": {"nl": False, "el": False, "na": False, "ea": False,
                         "link": "https://ls11-www.cs.uni-dortmund.de/peop" +
                         "le/morris/graphkerneldatasets/REDDIT-MULTI-12K.zip"},
    "SYNTHETIC": {"nl": False, "el": False, "na": True, "ea": False,
                  "link": "https://ls11-www.cs.uni-dortmund.de/people" +
                  "/morris/graphkerneldatasets/SYNTHETIC.zip"},
    "SYNTHETICnew": {"nl": False, "el": False, "na": True, "ea": False,
                     "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                     "morris/graphkerneldatasets/SYNTHETICnew.zip"},
    "Synthie": {"nl": False, "el": False, "na": True, "ea": False,
                "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                "morris/graphkerneldatasets/Synthie.zip"},
    "Tox21_AHR": {"nl": True, "el": True, "na": False, "ea": False,
                  "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                  "morris/graphkerneldatasets/Tox21_AHR.zip"},
    "Tox21_AR": {"nl": True, "el": True, "na": False, "ea": False,
                 "link": "https://ls11-www.cs.uni-dortmund.de/people/morris/" +
                 "graphkerneldatasets/COX2_MD.zip"},
    "Tox21_AR-LBD": {"nl": True, "el": True, "na": False, "ea": False,
                     "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                     "morris/graphkerneldatasets/Tox21_AR-LBD.zip"},
    "Tox21_ARE": {"nl": True, "el": True, "na": False, "ea": False,
                  "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                  "morris/graphkerneldatasets/Tox21_ARE.zip"},
    "Tox21_aromatase": {"nl": True, "el": True, "na": False, "ea": False,
                        "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                        "morris/graphkerneldatasets/Tox21_aromatase.zip"},
    "Tox21_ATAD5": {"nl": True, "el": True, "na": False, "ea": False,
                    "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                    "morris/graphkerneldatasets/Tox21_ATAD5.zip"},
    "Tox21_ER": {"nl": True, "el": True, "na": False, "ea": False,
                 "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                 "morris/graphkerneldatasets/Tox21_ER.zip"},
    "Tox21_ER_LBD": {"nl": True, "el": True, "na": False, "ea": False,
                     "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                     "morris/graphkerneldatasets/Tox21_ER_LBD.zipp"},
    "Tox21_HSE": {"nl": True, "el": True, "na": False, "ea": False,
                  "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                  "morris/graphkerneldatasets/Tox21_HSE.zip"},
    "Tox21_MMP": {"nl": True, "el": True, "na": False, "ea": False,
                  "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                  "morris/graphkerneldatasets/Tox21_MMP.zip"},
    "Tox21_p53": {"nl": True, "el": True, "na": False, "ea": False,
                  "link": "https://ls11-www.cs.uni-dortmund.de/people/" +
                  "morris/graphkerneldatasets/Tox21_p53.zip"},
    "Tox21_PPAR-gamma": {"nl": True, "el": True, "na": False, "ea": False,
                         "link": "https://ls11-www.cs.uni-dortmund.de/peop" +
                         "le/morris/graphkerneldatasets/Tox21_PPAR-gamma.zip"}
}

symmetric_dataset = False


def read_data(
        name,
        with_classes=True,
        prefer_attr_nodes=False,
        prefer_attr_edges=False,
        produce_labels_nodes=False,
        as_graphs=False,
        is_symmetric=symmetric_dataset):
    """Create a dataset iterable for GraphKernel.

    Parameters
    ----------
    name : str
        The dataset name.

    with_classes : bool, default=False
        Return an iterable of class labels based on the enumeration.

    produce_labels_nodes : bool, default=False
        Produce labels for nodes if not found.
        Currently this means labeling its node by its degree inside the Graph.
        This operation is applied only if node labels are non existent.

    prefer_attr_nodes : bool, default=False
        If a dataset has both *node* labels and *node* attributes
        set as labels for the graph object for *nodes* the attributes.

    prefer_attr_edges : bool, default=False
        If a dataset has both *edge* labels and *edge* attributes
        set as labels for the graph object for *edge* the attributes.

    as_graphs : bool, default=False
        Return data as a list of Graph Objects.

    is_symmetric : bool, default=False
        Defines if the graph data describe a symmetric graph.

    Returns
    -------
    Gs : iterable
        An iterable of graphs consisting of a dictionary, node
        labels and edge labels for each graph.

    classes : np.array, case_of_appearance=with_classes==True
        An one dimensional array of graph classes aligned with the lines
        of the `Gs` iterable. Useful for classification.

    """
    indicator_path = "./"+str(name)+"/"+str(name)+"_graph_indicator.txt"
    edges_path = "./" + str(name) + "/" + str(name) + "_A.txt"
    node_labels_path = "./" + str(name) + "/" + str(name) + "_node_labels.txt"
    node_attributes_path = "./"+str(name)+"/"+str(name)+"_node_attributes.txt"
    edge_labels_path = "./" + str(name) + "/" + str(name) + "_edge_labels.txt"
    edge_attributes_path = \
        "./" + str(name) + "/" + str(name) + "_edge_attributes.txt"
    graph_classes_path = \
        "./" + str(name) + "/" + str(name) + "_graph_labels.txt"

    # node graph correspondence
    ngc = dict()
    # edge line correspondence
    elc = dict()
    # dictionary that keeps sets of edges
    Graphs = dict()
    # dictionary of labels for nodes
    node_labels = dict()
    # dictionary of labels for edges
    edge_labels = dict()

    # Associate graphs nodes with indexes
    with open(indicator_path, "r") as f:
        for (i, line) in enumerate(f, 1):
            ngc[i] = int(line[:-1])
            if int(line[:-1]) not in Graphs:
                Graphs[int(line[:-1])] = set()
            if int(line[:-1]) not in node_labels:
                node_labels[int(line[:-1])] = dict()
            if int(line[:-1]) not in edge_labels:
                edge_labels[int(line[:-1])] = dict()

    # Extract graph edges
    with open(edges_path, "r") as f:
        for (i, line) in enumerate(f, 1):
            edge = line[:-1].replace(' ', '').split(",")
            elc[i] = (int(edge[0]), int(edge[1]))
            Graphs[ngc[int(edge[0])]].add((int(edge[0]), int(edge[1])))
            if is_symmetric:
                Graphs[ngc[int(edge[1])]].add((int(edge[1]), int(edge[0])))

    # Extract node attributes
    if (prefer_attr_nodes and
        dataset_metadata[name].get(
                "na",
                os.path.exists(node_attributes_path)
                )):
        with open(node_attributes_path, "r") as f:
            for (i, line) in enumerate(f, 1):
                node_labels[ngc[i]][i] = \
                    [float(num) for num in
                     line[:-1].replace(' ', '').split(",")]
    # Extract node labels
    elif dataset_metadata[name].get(
            "nl",
            os.path.exists(node_labels_path)
            ):
        with open(node_labels_path, "r") as f:
            for (i, line) in enumerate(f, 1):
                node_labels[ngc[i]][i] = int(line[:-1])
    elif produce_labels_nodes:
        for i in range(1, len(Graphs)+1):
            node_labels[i] = dict(Counter(s for (s, d) in Graphs[i] if s != d))

    # Extract edge attributes
    if (prefer_attr_edges and
        dataset_metadata[name].get(
            "ea",
            os.path.exists(edge_attributes_path)
            )):
        with open(edge_attributes_path, "r") as f:
            for (i, line) in enumerate(f, 1):
                attrs = [float(num)
                         for num in line[:-1].replace(' ', '').split(",")]
                edge_labels[ngc[elc[i][0]]][elc[i]] = attrs
                if is_symmetric:
                    edge_labels[ngc[elc[i][1]]][(elc[i][1], elc[i][0])] = attrs

    # Extract edge labels
    elif dataset_metadata[name].get(
            "el",
            os.path.exists(edge_labels_path)
            ):
        with open(edge_labels_path, "r") as f:
            for (i, line) in enumerate(f, 1):
                edge_labels[ngc[elc[i][0]]][elc[i]] = int(line[:-1])
                if is_symmetric:
                    edge_labels[ngc[elc[i][1]]][(elc[i][1], elc[i][0])] = \
                        int(line[:-1])

    Gs = list()
    if as_graphs:
        for i in range(1, len(Graphs)+1):
            Gs.append(Graph(Graphs[i], node_labels[i], edge_labels[i]))
    else:
        for i in range(1, len(Graphs)+1):
            Gs.append([Graphs[i], node_labels[i], edge_labels[i]])

    if with_classes:
        classes = []
        with open(graph_classes_path, "r") as f:
            for line in f:
                classes.append(int(line[:-1]))

        classes = np.array(classes, dtype=np.int)
        return Bunch(data=Gs, target=classes)
    else:
        return Bunch(data=Gs)


def _download_zip(url, output_name):
    """Download a file from a requested url and store locally.

    Parameters
    ----------
    url : str
        The url from where the file will be downloaded.

    output_name : str
        The name of the file in the local directory.

    Returns
    -------
    None.

    """
    ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    filename = output_name + ".zip"
    try:
        data_url = urlopen(url, context=ctx)
    except HTTPError as e:
        if e.code == 404:
            e.msg = "Dataset '%s' not found on mldata.org." % output_name
        raise

    # Store Zip File
    try:
        with open(filename, 'w+b') as zip_file:
            copyfileobj(data_url, zip_file)
    except Exception:
        os.remove(filename)
        raise
    data_url.close()


def fetch_dataset(
        name,
        verbose=True,
        data_home=None,
        download_if_missing=True,
        with_classes=True,
        produce_labels_nodes=False,
        prefer_attr_nodes=False,
        prefer_attr_edges=False,
        as_graphs=False):
    """Access a large collection of benchmark datasets from TU Dortmund :cite:`KKMMN2016`.

    For more info visit: :xref:`gd`

    Parameters
    ----------
    name : str
        The name of the dataset (as found in :xref:`gd`).

    verbose : bool, default=True
        Print messages, throughout execution.

    data_home : string, default=None
        Specify another download and cache folder for the datasets.
        By default all grakel data is stored in '~/grakel_data' subfolders.

    download_if_missing : boolean, default=True
        If False, raise a IOError if the data is not locally available instead
        of trying to download the data from the source site.

    with_classes : bool, default=False
        Return an iterable of class labels based on the enumeration.

    produce_labels_nodes : bool, default=False
        Produce labels for nodes if not found.
        Currently this means labeling its node by its degree inside the Graph.
        This operation is applied only if node labels are non existent.

    prefer_attr_nodes : bool, default=False
        If a dataset has both *node* labels and *node* attributes
        set as labels for the graph object for *nodes* the attributes.

    prefer_attr_edges : bool, default=False
        If a dataset has both *edge* labels and *edge* attributes
        set as labels for the graph object for *edge* the attributes.

    as_graphs : bool, default=False
        Return data as a list of Graph Objects.

    Returns
    -------
    graphs : iterable
        Returns an iterable of the produced *valid-graph-format*
        and labels for each node.

    classes : list
        Returns a list of all the classes corresponding to each graph by
        order of input.

    """
    name = str(name)
    if name in dataset_metadata:
        if data_home is None:
            data_home = os.path.join(os.path.expanduser("~"), 'grakel_data')

        exists = os.path.isdir(data_home)
        missing = not os.path.exists(os.path.join(data_home, name + ".zip"))
        cwd = os.getcwd()

        if missing:
            if download_if_missing:
                if not exists:
                    if verbose:
                        print("Initializing folder at", str(data_home))
                    os.makedirs(data_home)
                os.chdir(data_home)
                if verbose:
                    print("Downloading dataset for", name + "..")
                _download_zip(dataset_metadata[name]["link"], name)
            else:
                raise IOError('Dataset ' + name +
                              ' was not found on ' + str(data_home))
        else:
            # move to the general data directory
            os.chdir(data_home)

        with zipfile.ZipFile(str(name) + '.zip', "r") as zip_ref:
            if verbose:
                print("Extracting dataset ", str(name) + "..")
            zip_ref.extractall()

        if verbose:
            print("Parsing dataset ", str(name) + "..")

        data = read_data(name,
                         with_classes=with_classes,
                         prefer_attr_nodes=prefer_attr_nodes,
                         prefer_attr_edges=prefer_attr_edges,
                         produce_labels_nodes=produce_labels_nodes,
                         is_symmetric=symmetric_dataset,
                         as_graphs=as_graphs)
        if verbose:
            print("Parse was succesful..")

        if verbose:
            print("Deleting unzipped dataset files..")
        shutil.rmtree(str(name))

        if verbose:
            print("Going back to the original directory..")
        os.chdir(cwd)

        return data
    else:
        raise ValueError('Dataset: "'+str(name)+'" is currently unsupported.' +
                         '\nSupported datasets come from '
                         'https://ls11-www.cs.tu-dortmund.de/staff/morris/' +
                         'graphkerneldatasets. If your dataset name appears' +
                         ' them send us a pm, to explain you either why we ' +
                         'don\'t support it, or to update our dataset ' +
                         'database.')


def get_dataset_info(dataset_name, default=None):
    """Return the info concerning the existence of a certain dataset.

    Parameters
    ----------
    dataset_name : str
        The name of the dataset.

    default : Object, default=None
        The default return value if the dataset is not found.

    Returns
    -------
    dictionary_get : default or dictionary_entry
        The info of a dataset as a dictionary with
        fields:

            - **nl** : A boolean flag indicating if the
              dataset has node labels.

            - **el** : A boolean flag indicating if the
              dataset has edge labels.

            - **na** : A boolean flag indicating if the
              dataset has node attributes.

            - **ea** : A boolean flag indicating if the
              dataset has edge attributes.

            - **link** : A str corresponding to the
              download link of the dataset.

    """
    return dataset_metadata.get(dataset_name, default)



"""A file containing useful external functions"""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
import os
import numpy as np

from collections import defaultdict
from collections import Iterable

from sklearn.base import TransformerMixin
from sklearn.pipeline import make_pipeline
from sklearn.model_selection import KFold
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import ShuffleSplit
from sklearn.base import BaseEstimator
from sklearn.svm import SVC
from sklearn.utils import Bunch
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_is_fitted

from grakel import Graph
from grakel.graph import is_adjacency as valid_matrix


class KMTransformer(BaseEstimator, TransformerMixin):
    """A Kernel Matrix Transformer.

    Usefull for using precalculated Kernel Matrices inside scikit-learn pipeline.

    Parameters
    ----------
    K : array-like, shape=[n, n]
        If given an array the input can be as follows:

            + array-like lists of lists

            + np.array

            + sparse matrix (scipy.sparse)

        It can also be embedded in an sklearn Bunch object as a mat (argument)

    Attributes
    ----------
    K_ : numpy.array, shape=[n, n]

    """
    def __init__(self, K=None):
        """Initialise the Kernel Matrix Transformer"""

        self.K = K
        self._initialized = {"K": False}

    def initialize(self):
        """Initialize all transformer arguments, needing initialisation."""
        if not self._initialized["K"]:
            if self.K is None:
                M = np.array([[1.0]])
            else:
                K = self.K
                if isinstance(K, Bunch):
                    try:
                        K = K.mat
                    except Exception:
                        raise ValueError('If in an sklearn Bunch K must be under mat')
                flag, M = valid_matrix(K, transform=True)
                if not flag:
                    raise ValueError('The provided K cannot be converted to a '
                                     'two dimensional np.array.')
            self.K_ = M
            self._initialized["K"] = True

    def fit(self, X, y=None):
        """Fit a list of indeces.

        Parameters
        ----------
        X : Iterable of int:
            Indexes for the X of the kernel matrix.

        Returns
        -------
        self : object
            Returns self.

        """
        self.initialize()
        if any(x < 0 or x > self.K_.shape[0] for x in X):
            raise ValueError('')
        else:
            self.X = np.array(X)

        return self

    def fit_transform(self, X, y=None):
        """Fit and transform, on the same dataset.

        Parameters
        ----------
        X : Iterable of int
            Indexes for the first dimension of the kernel matrix.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        K : numpy array, shape = [len(X), len(X)]
            Corresping to the values of the X indexes with themselfs.

        """
        # Initialize the Graph Kernel
        self.initialize()
        if any(x < 0 or x > self.K_.shape[0] for x in X):
            raise ValueError('')
        else:
            self.X = np.array(X)

        return self.K_[self.X, :][:, self.X]

    def transform(self, X):
        """Calculate the kernel matrix, between given and fitted dataset.

        Parameters
        ----------
        X : Iterable of int
            Indexes for the second dimension of the kernel matrix.

        Returns
        -------
        K : numpy array, shape = [len(Y), len(X)]
            Corresping to the values of the Y indexes with X.

        """
        check_is_fitted(self, 'X')
        if any(x < 0 or x > self.K_.shape[0] for x in X):
            raise ValueError('')

        return self.K_[X, :][:, self.X]


def cross_validate_Kfold_SVM(K, y,
                             n_iter=10, n_splits=10, C_grid=None,
                             random_state=None, scoring="accuracy", fold_reduce=None):
    """Cross Validate a list of precomputed kernels with an SVM.

    Parameters
    ----------
    K : list
        A list that must contain either numpy arrays or iterables of numpy arrays.

    y : list
        List of lists that for every element of K contains numbers of score for all iterations.

    n_iter : int
        Number of iteration for the K-Fold.

    n_splits : int
        Number of splits for the K-Fold.

    random_state :  RandomState or int, default=None
        A random number generator instance or an int to initialize a RandomState as a seed.

    fold_reduce : callable or None
        A function that summarizes information between all folds.
        Input must be a list of n_splits elements corresponding to scoring.
        If None default is np.mean.

    scoring : string, callable, list/tuple, dict or None, default: None
        As in ``scoring`` in :xref:`lgscv`.

    Returns
    -------
    out : list
       A list that contains a list of the reduced folds for each iteration, for each primary
       element of K.
    """
    # Initialise C_grid
    if C_grid is None:
        C_grid = ((10. ** np.arange(-7, 7, 2)) / len(y)).tolist()
    elif type(C_grid) is np.array:
        C_grid = np.squeeze(C_grid)
        if len(C_grid.shape) != 1:
            raise ValueError('C_grid should either be None or a squeezable to 1 dimension np.array')
        else:
            C_grid = list(C_grid)

    # Initialise fold_reduce:
    if fold_reduce is None:
        fold_reduce = np.mean
    elif not isinstance(callable, fold_reduce):
        raise ValueError('fold_reduce should be a callable')

    # Initialise and check random state
    random_state = check_random_state(random_state)

    # Initialise sklearn pipeline objects
    kfolder = KFold(n_splits=n_splits, random_state=random_state, shuffle=True)
    estimator = make_pipeline(KMTransformer(), SVC(kernel='precomputed'))

    # Make all the requested folds
    nfolds = tuple(tuple(kfolder.split(y)) for _ in range(n_iter))

    out = list()
    for ks in K:
        mid = list()
        if valid_matrix(ks):
            pg = {"svc__C": C_grid, "kmtransformer__K": [Bunch(mat=ks)]}
        elif isinstance(ks, Iterable) and all(valid_matrix(k) for k in ks):
            pg = [{"svc__C": C_grid, "kmtransformer__K": [Bunch(mat=k)]} for k in ks]
        else:
            raise ValueError('Not a valid object for kernel matrix/ces')

        for kfolds in nfolds:
            fold_info = list()
            for train, test in kfolds:
                gs = GridSearchCV(estimator, param_grid=pg, scoring=scoring,
                                  cv=ShuffleSplit(n_splits=1,
                                                  test_size=0.1,
                                                  random_state=random_state)).fit(train, y[train])
                fold_info.append(gs.score(test, y[test]))
            mid.append(fold_reduce(fold_info))
        out.append(mid)
    return out


def graph_from_networkx(X, node_labels_tag=None, edge_labels_tag=None, edge_weight_tag=None,
                        as_Graph=False):
    """Transform an iterable of networkx objects to an iterable of Graphs.

    A function for helping a user that has a collection of graphs in networkx to use grakel.

    Parameters
    ----------
        node_labels_tag : str or None
            Define where to search for labels of nodes, inside the `node` attribute of each graph.
            If None no labels are assigned.

        edge_labels_tag : tuple
            Define where to search for labels of edges, inside the `edge` attribute of each graph.
            If None no labels are assigned.

        edge_weight_tag : int
            Define where to search for weights inside the `edge` attribute of each graph.
            If None 1.0 weights are assigned.

        as_Graph : bool, default=False
            Return each output as a grakel.Graph object.

    Returns
    -------
        grakel_graphs : generator
            Returns a generator containing the same collections to be given as an input for
            any grakel kernel.

    """
    import networkx as nx
    v2 = False
    if nx.__version__ > '2.':
        v2 = True

    if node_labels_tag is None:
        def nodel_init():
            return None

        def nodel_put(*args):
            pass
    elif type(node_labels_tag) is str:
        def nodel_init():
            return dict()

        def nodel_put(nl, u, d):
            nl[u] = d[u][node_labels_tag]
    else:
        raise ValueError('node_labels_tag must be a str indicating the '
                         'tag of the labels inside nodes or None')

    if edge_labels_tag is None:
        def edgel_init():
            return None

        def edgel_put(*args):
            pass
    elif type(edge_labels_tag) is str:
        def edgel_init():
            return dict()

        if v2:
            def edgel_put(el, u, d):
                el[u] = d[u][edge_labels_tag]
        else:
            def edgel_put(el, u, d):
                el[u] = d[u[0]][u[1]][edge_labels_tag]
    else:
        raise ValueError('edge_labels_tag must be a str indicating the '
                         'tag of the labels inside edges or None')

    if edge_weight_tag is None:
        def get_weight(*args):
            return 1.0
    elif type(edge_weight_tag) is str:
        if v2:
            def get_weight(d, e):
                return d[e][edge_weight_tag]
        else:
            def get_weight(d, e):
                return d[e[0]][e[1]][edge_weight_tag]

    else:
        raise ValueError('weight_labels_tag must be a str indicating  '
                         'tag of the labels inside edges or None (1.0)')

    if not isinstance(X, Iterable):
        raise ValueError('X must be an iterable')

    if v2:
        def take_ne(graph):
            return graph.nodes, graph.edges
    else:
        def take_ne(graph):
            return graph.node, graph.edge

    for G in X:
        graph_object = dict()
        nl = nodel_init()
        el = edgel_init()
        nodes, edges = take_ne(G)
        for u in G.nodes():
            graph_object[u] = dict()
            nodel_put(nl, u, nodes)
            for v in G.neighbors(u):
                graph_object[u][v] = get_weight(edges, (u, v))
                edgel_put(el, (u, v), edges)

        if as_Graph:
            yield Graph(graph_object, nl, el)
        else:
            yield [graph_object, nl, el]


def graph_from_pandas(edge_df, node_df=None, directed=False, as_Graph=False):
    """Produces a collection of Graph Objects from pandas dataframes.

    A function for helping a user that has a bunch of graph csv
    files make a collection for grakel input.

    Parameters
    ----------
        edge_df : tuple(pandas.DataFrame, pct, tuple(pct, pct), pct or None, pct or list(pct) or None)
            A tuple of many elements:

                1. A pandas dataframe containing all the edges of a collection of graphs
                2. The column name of the graph index
                3. A tuple containing source and destination column names
                4. The column name of weights column (None if non-existent)
                5. The column name pointing the column containing edge-labels or a list pointing
                   the columns of attributes (None if edge-labels are non-existent).

            If node_df exists, correspondance of graph indexes and node indexes must be exact.

        node_df : tuple(pandas.DataFrame, pct, pct or list(pct) or None) or None, default=None
            A tuple of many elements:

                1. A pandas dataframe containing all the nodes of a collection of graphs
                2. The column name of the graph index
                3. The column name pointing the column containing node-labels or a list pointing
                   the columns of attributes (None if node-labels are non-existent).

            Node id must correspond to node number.

        directed : bool, default=False
            A variable indicating with True if both directions of the graph have been inserted inside
            the dataframe. Otherwise the graph is considered as undirected.

        as_Graph : bool, default=False
            Return each output as a grakel.Graph object.

    Returns
    -------
        grakel_graphs : dict
            Dictionary with keys te keys of the graphs inside the pandas dataframe and values the
            corresponding graph object

    """
    from pandas import DataFrame

    if node_df is None:
        graphs = dict()
        throw_error = False
    elif (type(node_df) is tuple and
          len(node_df) == 3 and
          type(node_df[0]) is DataFrame and
          (node_df[1] is None or node_df[1] in node_df[0].columns) and
          (node_df[2] is None or
           (type(node_df[2]) is not list and node_df[2] in node_df[0].columns) or
           (type(node_df[2]) is list and all(i in node_df[0].columns for i in node_df[2])))):
        df, gtag, labs = node_df
        graphs = defaultdict(lambda: defaultdict(dict))
        if labs is None:
            for index, row in df.iterrows():
                gidx = row[gtag]
                graphs[gidx]["graph"][index] = dict()
                graphs[gidx]["node_label"] = None
        elif type(labs) is list:
            for index, row in df.iterrows():
                gidx = row[gtag]
                graphs[gidx]["graph"][index] = dict()
                graphs[gidx]["node_label"][index] = np.array(row[c] for c in labs)
        else:
            for index, row in df.iterrows():
                gidx = row[gtag]
                graphs[gidx]["graph"][index] = dict()
                graphs[gidx]["node_label"][index] = row[labs]
        throw_error = True
    else:
        raise ValueError('node_df must be a tuple containing a pandas.DataFrame object '
                         'a column name corresponding to the valid column index for graphs indexes '
                         'and a None, column name or list of column names corresponding to no-labels, '
                         'labels or attributes.')

    if (type(edge_df) is tuple and
        len(edge_df) == 5 and
        type(edge_df[0]) is DataFrame and
        (edge_df[1] in edge_df[0].columns) and
        (type(edge_df[2]) is tuple and len(edge_df[2]) == 2 and
         all(c in edge_df[0].columns for c in edge_df[2])) and
        (edge_df[3] is None or edge_df[3] in edge_df[0].columns) and
        (edge_df[4] is None or
         (type(edge_df[4]) is not list and edge_df[4] in edge_df[0].columns) or
         (type(edge_df[4]) is list and all(c in edge_df[0].columns for c in edge_df[4])))):
        df, gtag, (src_c, dst_c), w_c, labs = edge_df
        if w_c is not None:
            def get_weight(row):
                return row[w_c]
        else:
            def get_weight(row):
                return 1.

        if labs is None:
            def set_label(*args):
                pass
        elif type(labs) is list:
            def set_label(row, d, e):
                if d["edge_label"] is None:
                    d["edge_label"] = dict()
                el = np.array([row[l] for l in labs])
                d["edge_label"][e] = el
                if not directed:
                    d["edge_label"][(e[1], e[0])] = el
        else:
            def set_label(row, d, e):
                if d["edge_label"] is None:
                    d["edge_label"] = dict()
                el = row[labs]
                d["edge_label"][e] = el
                if not directed:
                    d["edge_label"][(e[1], e[0])] = el

        for index, row in df.iterrows():
            src, dst = row[src_c], row[dst_c]
            gidx = row[gtag]
            if gidx not in graphs:
                if throw_error:
                    raise ValueError('This graph didn\'t appear on node labels dataframe')
                else:
                    graphs[gidx] = dict()
                    graphs[gidx]["graph"] = dict()
                    if "node_label" not in graphs[gidx]:
                        graphs[gidx]["node_label"] = None
                    graphs[gidx]["edge_label"] = None
            if src not in graphs[gidx]["graph"]:
                if throw_error:
                    raise ValueError('This node didn\'t appear on node labels dataframe for graph '
                                     'with id ' + str(gidx))
                else:
                    graphs[gidx]["graph"][src] = dict()
            graphs[gidx]["graph"][src][dst] = get_weight(row)
            if not directed:
                if dst not in graphs[gidx]["graph"]:
                    if throw_error:
                        raise ValueError('This node didn\'t appear on node labels dataframe '
                                         'for graph with id ' + str(gidx))
                    else:
                        graphs[gidx]["graph"][dst] = dict()
                graphs[gidx]["graph"][dst][src] = get_weight(row)
            set_label(row, graphs[gidx], (src, dst))
    else:
        raise ValueError('edge_df must be a tuple containing a pandas.DataFrame object '
                         'a column name corresponding to a valid column index of the dataframe '
                         'that contain the index of the graph that an edge belongs inside the '
                         'dataframe, a column name corresponding to weights inside the dataframe'
                         '(or None if weights do not exist) and a list of column names '
                         'corresponding or column name corresponding to labels or None '
                         'if none of this exists')

    return {k: Graph(graphs[k]["graph"], graphs[k]["node_label"], graphs[k]["edge_label"])
            if as_Graph else [graphs[k]["graph"], graphs[k]["node_label"], graphs[k]["edge_label"]]
            for k in graphs.keys()}


def graph_from_csv(edge_files, node_files=None, index_type=str,
                   directed=False, sep=",", as_Graph=False):
    """Produces a collection of Graph Objects from a collection of csv files.

    A function for helping a user that has a bunch of graph csv
    files make a collection for grakel input.

    Parameters
    ----------
        edge_files : tuple(iter(str), bool, bool or None)
            edge_files = (iter(<edge_file_address>), <weight_flag>, <attributes_flag>).
            Each line of all the edge files must have the following structure:

            .. code::

                <v_i><sep><v_j>[weight][labels/attr]

                weights: "<sep><weight>", <weight_flag> is True
                         "", <weight_flag> is False
                labels/attr: "<sep><label>", <attributes_flag> is False
                             "<sep><a_1>, .. , <sep><a_n>", <attributes_flag> is True
                             "", <attributes_flag> is None

        node_files : tuple(iter(str), bool) or None, default=None
            node_files = (iter(<node_file_address>), index_type, <attributes_flag>).
            Each line of the node file must have the following structure (if exists):

            .. code::

                <v_i><sep>[labels/attr]

                labels/attr: "<sep><label>", <attributes_flag> is False
                             "<sep><a_1>, .. , <sep><a_n>", <attributes_flag> is True
                             "", <attributes_flag> is None

        directed : bool, default=False
            Defines if the graph should be considered as directed.

        sep : str, default=","
            The separator for the csv files.

        as_Graph : bool, default=False
            Return each output as a grakel.Graph object.

    Returns
    -------
        grakel_graphs : generator
            Returns a generator containing the same collections to be given as an input for
            any grakel kernel.

    """
    def edge_files_error():
        raise ValueError('edge_file argument must contain an iterable of strings of edge files, '
                         'a bool weight_flag and attributes_flag bool or None')

    def node_files_error():
        raise ValueError('node_files argument can be None or contain an iterable of strings'
                         'of edge files and attributes_flag bool or None')

    if type(index_type) is not type:
        raise ValueError('index_type must be a class `type` object')

    if (type(edge_files) is not tuple or len(edge_files) != 3 or
            type(edge_files[1]) is not bool or
            type(edge_files[2]) not in [bool, None]):
        edge_files_error()
    else:
        if edge_files[1]:
            def get_weight(l):
                return float(l.pop(0))
        else:
            def get_weight(l):
                return 1.0
        efs = list()
        for e in edge_files[0]:
            if type(e) is not str:
                edge_files_error()
            if not os.path.isfile(e):
                raise ValueError('Each edge file address must be a valid address')
            efs.append(e)

    if type(node_files) is None:
        nfs = None
    elif (type(node_files) is tuple and len(node_files) == 2 and
            type(node_files[1]) in [bool, None]):
        nfs = list()
        for n in node_files[0]:
            if type(n) is not str:
                edge_files_error()
            if not os.path.isfile(n):
                raise ValueError('Each edge file address must be a valid address')
            nfs.append(n)
    else:
        node_files_error()

    if edge_files[2] is None:
        def get_el(ef, g):
            el = None
            for line in ef:
                q = line.split(sep)
                m = q[2:]
                ea, eb = [index_type(e) for e in q[:2]]
                g[ea][eb] = get_weight(m)
                if not directed:
                    g[eb][ea] = g[ea][eb]
            return el
    elif edge_files[2]:
        def get_el(ef, g):
            el = dict()
            for line in ef:
                q = line.strip('\n').split(sep)
                ea, eb = [index_type(e) for e in q[:2]]
                m = q[2:]
                g[ea][eb] = get_weight(m)
                el[(ea, eb)] = np.array([float(num) for num in m])
                if not directed:
                    el[(eb, ea)] = el[(ea, eb)]
                    g[eb][ea] = g[ea][eb]
            return el
    else:
        def get_el(ef, g):
            el = dict()
            for line in ef:
                q = line.strip('\n').split(sep)
                ea, eb = [index_type(e) for e in q[:2]]
                m = q[2:]
                g[ea][eb] = get_weight(m)
                el[(ea, eb)] = str(m[0])
                if not directed:
                    el[(eb, ea)] = el[(ea, eb)]
                    g[eb][ea] = g[ea][eb]
            return el

    if nfs is None:
        for efa in efs:
            nl = None
            graph_object = defaultdict(dict)
            with open(efa, 'r') as ef:
                el = get_el(ef, graph_object)
            graph_object = dict(graph_object)

            if as_Graph:
                yield Graph(graph_object, nl, el)
            else:
                yield [graph_object, nl, el]
    else:
        if len(nfs) != len(efs):
            raise ValueError('The number of edge files and the number of node files must be the same')
        for (nfa, efa) in zip(nfs, efs):
            graph_object = dict()
            with open(nfa, 'r') as nf:
                if node_files[1] is None:
                    nl = None
                    for line in nf:
                        graph_object[index_type(line.strip('\n').split(sep)[0].strip())] = dict()
                elif node_files[1]:
                    nl = dict()
                    for line in nf:
                        q = line.strip('\n').split(sep)
                        graph_object[index_type(q[0].strip())] = dict()
                        nl[index_type(q[0].strip())] = np.array([float(i.strip()) for i in q[1:]])
                else:
                    nl = dict()
                    for line in nf:
                        q = line.strip('\n').split(sep)
                        graph_object[index_type(q[0].strip())] = dict()
                        nl[index_type(q[0].strip())] = q[1]

            with open(efa, 'r') as ef:
                el = get_el(ef, graph_object)

            if as_Graph:
                yield Graph(graph_object, nl, el)
            else:
                yield [graph_object, nl, el]



"""Tests for utils.py"""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
from __future__ import print_function

import os

from warnings import warn
from numpy import arange
from numpy.random import RandomState
from grakel import graph_from_pandas
from grakel import graph_from_networkx
from grakel import graph_from_csv
from grakel import cross_validate_Kfold_SVM
from grakel import ShortestPath
from grakel import WeisfeilerLehman
from grakel import VertexHistogram
from grakel.datasets import fetch_dataset
from grakel.datasets.base import read_data

global verbose

# Add extra arguments for allowing unit testing
if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(
        description='A test file for all `Graph` type objects')
    parser.add_argument(
        '--verbose',
        help='verbose outputs on stdout',
        action="store_true")
    parser.add_argument(
        '--ignore_warnings',
        help='ignore warnings produced by kernel executions',
        action="store_true")

    args = parser.parse_args()
    verbose = bool(args.verbose)

    if bool(args.ignore_warnings):
        import warnings
        warnings.filterwarnings('ignore', category=UserWarning)
else:
    import warnings
    warnings.filterwarnings('ignore', category=UserWarning)
    verbose = False


def test_pandas():
    """Testing Graph object consistency for an adjacency-type initialization object."""
    # Input

    try:
        import pandas as pd
    except ImportError:
        return

    ga = {0: {0: 1., 1: 1., 3: 3.},
          1: {0: 1., 3: 2.},
          2: {0: 2., 1: 3., 3: 1.},
          3: {0: 1.}}

    ganl = {0: 'l1', 1: 'l2', 2: 'l3', 3: 'l4'}

    gael = {(0, 0): 'el1', (0, 1): 'el2', (0, 3): 'el3',
            (1, 0): 'el4', (1, 3): 'el5', (2, 0): 'el6',
            (2, 1): 'el7', (2, 3): 'el8', (3, 0): 'el9'}

    gb = {4: {4: 1., 5: 1., 7: 3.},
          5: {4: 1., 7: 2.},
          6: {4: 2., 5: 3., 7: 1.},
          7: {4: 1.}}

    gbnl = {4: 'l1', 5: 'l2', 6: 'l3', 7: 'l4'}

    gbel = {(4, 4): 'el1', (4, 5): 'el2', (4, 7): 'el3',
            (5, 4): 'el4', (5, 7): 'el5', (6, 4): 'el6',
            (6, 5): 'el7', (6, 7): 'el8', (7, 4): 'el9'}

    nodes_df = pd.DataFrame({'l': ['l1', 'l2', 'l3', 'l4', 'l1', 'l2', 'l3', 'l4'],
                             'g': [0, 0, 0, 0, 1, 1, 1, 1]})

    edges_df = pd.DataFrame({'src': [0, 0, 0, 1, 1, 2, 2, 2, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7],
                             'dst': [0, 1, 3, 0, 3, 0, 1, 3, 0, 4, 5, 7, 4, 7, 4, 5, 7, 4],
                             'w': [1., 1., 3., 1., 2., 2., 3., 1., 1.,
                                   1., 1., 3., 1., 2., 2., 3., 1., 1.],
                             'g': [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                             'l': ['el1', 'el2', 'el3', 'el4', 'el5', 'el6', 'el7', 'el8', 'el9',
                                   'el1', 'el2', 'el3', 'el4', 'el5', 'el6', 'el7', 'el8', 'el9']})
    gs = graph_from_pandas((edges_df, 'g', ('src', 'dst'), 'w', 'l'),
                           (nodes_df, 'g', 'l'), directed=True)

    assert(gs[0][0] == ga and gs[0][1] == ganl and gs[0][2] == gael and
           gs[1][0] == gb and gs[1][1] == gbnl and gs[1][2] == gbel)


def test_networkx():
    """Testing Graph object consistency for an edge-dictionary-type initialization object."""

    try:
        import networkx as nx
    except ImportError:
        return

    ga = {0: {0: 1., 1: 1., 3: 3.},
          1: {0: 1., 3: 2.},
          2: {0: 2., 1: 3., 3: 1.},
          3: {0: 1.}}

    ganl = {0: 'l1', 1: 'l2', 2: 'l3', 3: 'l4'}

    gael = {(0, 0): 'el1', (0, 1): 'el2', (0, 3): 'el3',
            (1, 0): 'el4', (1, 3): 'el5', (2, 0): 'el6',
            (2, 1): 'el7', (2, 3): 'el8', (3, 0): 'el9'}

    g = nx.DiGraph()

    for n in ganl.keys():
        g.add_node(n, nl=ganl[n])

    for e in gael.keys():
        g.add_edge(e[0], e[1], w=ga[e[0]][e[1]], el=gael[e])

    gs = list(graph_from_networkx([g], 'nl', 'el', 'w'))[0]

    assert(gs[0] == ga and gs[1] == ganl and gs[2] == gael)


def test_csv():
    """Testing Graph object consistency for an edge-dictionary-type initialization object."""

    ga = {0: {0: 1., 1: 1., 3: 3.},
          1: {0: 1., 3: 2.},
          2: {0: 2., 1: 3., 3: 1.},
          3: {0: 1.}}

    ganl = {0: 'l1', 1: 'l2', 2: 'l3', 3: 'l4'}

    gael = {(0, 0): 'el1', (0, 1): 'el2', (0, 3): 'el3',
            (1, 0): 'el4', (1, 3): 'el5', (2, 0): 'el6',
            (2, 1): 'el7', (2, 3): 'el8', (3, 0): 'el9'}

    node_files = ['temp1_n.csv', 'temp2_n.csv', 'temp3_n.csv']
    edge_files = ['temp1_e.csv', 'temp2_e.csv', 'temp3_e.csv']
    for (nf, ef) in zip(node_files, edge_files):
        with open(nf, 'w+') as f:
            for k in ganl.keys():
                print(str(k) + "," + ganl[k], file=f)
        with open(ef, 'w+') as f:
            for k in gael.keys():
                print(",".join(map(str, [k[0], k[1], ga[k[0]][k[1]], gael[k]])), file=f)

    graphs = list(graph_from_csv(edge_files=(edge_files, True, False), node_files=(node_files, False),
                                 index_type=int, directed=True))

    for f in node_files + edge_files:
        os.remove(f)

    assert(all(gs[0] == ga and gs[1] == ganl and gs[2] == gael for gs in graphs))


def test_KM_Kfold():
    """Testing KFold execution on wl, rw, sp input."""
    def load_mutag():
        try:
            dataset = fetch_dataset("MUTAG", with_classes=True, verbose=verbose)
        except Exception:
            # Offline testing
            warn('MUTAG could not be downloaded: using an offline version..')
            cwd = os.getcwd()
            os.chdir(os.path.join(os.path.dirname(__file__), 'data'))
            dataset = read_data('MUTAG', with_classes=True)
            os.chdir(cwd)
        return dataset.data, dataset.target

    # Input
    X, y = load_mutag()
    K = [ShortestPath(normalize=True).fit_transform(X),
         [WeisfeilerLehman(base_graph_kernel=VertexHistogram, n_iter=3, normalize=True).fit_transform(X),
          WeisfeilerLehman(base_graph_kernel=VertexHistogram, n_iter=5, normalize=True).fit_transform(X)]]

    # Parametrization
    n_splits = 10
    rs = RandomState(42)
    n_iter = 10
    C_grid = list(((10. ** arange(-7, 7, 2)) / len(y)).tolist())
    scoring = "accuracy"

    # Execute Kfold
    cross_validate_Kfold_SVM(K, y,
                             n_iter=n_iter, n_splits=n_splits, C_grid=C_grid,
                             random_state=rs, scoring=scoring, fold_reduce=None)


if __name__ == '__main__':
    test_pandas()
    test_networkx()
    test_csv()
    test_KM_Kfold()



"""Shortest path kernel as defined in :cite:`borgwardt2005shortest`."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
import collections
import warnings

import numpy as np

from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from grakel.graph import Graph
from grakel.kernels import Kernel


class ShortestPathAttr(Kernel):
    r"""The shortest path kernel for attributes.

    The Graph labels are considered as attributes.
    The computational efficiency is decreased to :math:`O(|V|^4)`
    See :cite:`borgwardt2005shortest`.

    Parameters
    ----------
    algorithm_type : str, default={"dijkstra", "floyd_warshall", "auto"}
        Apply the dijkstra or floyd_warshall algorithm for calculating
        shortest path, or chose automatically ("auto") based on the
        current graph format ("auto").

    metric : function, default=:math:`f(x,y)=\sum_{i}x_{i}*y_{i}`,
        The metric applied between attributes of the graph labels.
        The user must provide a metric based on the format of the provided
        labels (considered as attributes).

    Attributes
    ----------
    X : list
        A list of tuples, consisting of shortest path matrices
        and their feature vectors.

    """

    def __init__(self, n_jobs=None,
                 normalize=False,
                 verbose=False,
                 algorithm_type="auto",
                 metric=np.dot):
        """Initialise a `shortest_path_attr` kernel."""
        super(ShortestPathAttr, self).__init__(
            n_jobs=n_jobs, normalize=normalize, verbose=verbose)

        self.algorithm_type = algorithm_type
        self.metric = metric
        self._initialized.update({"algorithm_type": False, "metric": False})

    def initialize(self):
        """Initialize all transformer arguments, needing initialization."""
        super(ShortestPathAttr, self).initialize()
        if not self._initialized["algorithm_type"]:
            if self.algorithm_type == "auto":
                self._graph_format = "auto"
            elif self.algorithm_type == "floyd_warshall":
                self._graph_format = "adjacency"
            elif self.algorithm_type == "dijkstra":
                self._graph_format = "dictionary"
            else:
                raise ValueError('Unsupported value ' +
                                 str(self.algorithm_type) +
                                 ' for "algorithm_type"')
            self._initialized["algorithm_type"] = True

        if not self._initialized["metric"]:
            if not callable(self.metric):
                raise TypeError('"metric" must be callable')
            self._initialized["metric"] = True

    def parse_input(self, X):
        """Parse and create features for the `shortest_path` kernel.

        Parameters
        ----------
        X : iterable
            For the input to pass the test, we must have:
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that correspond to the given
            graph format). A valid input also consists of graph type objects.

        Returns
        -------
        sp_attr_tup : list
            A list of tuples of shortest path matrices and tehir attributes.

        """
        if not isinstance(X, collections.Iterable):
            raise TypeError('input must be an iterable\n')
        else:
            sp_attr_tup = list()
            ni = 0
            for (i, x) in enumerate(iter(X)):
                is_iter = isinstance(x, collections.Iterable)
                if is_iter:
                    x = list(x)
                if is_iter and len(x) in [0, 2, 3]:
                    if len(x) == 0:
                        warnings.warn('Ignoring empty element' +
                                      ' on index: '+str(i))
                        continue
                    else:
                        S, L = Graph(
                            x[0], x[1], {},
                            self._graph_format).build_shortest_path_matrix(
                                self.algorithm_type)
                elif type(x) is Graph:
                    S, L = x.build_shortest_path_matrix(self.algorithm_type)
                else:
                    raise TypeError('each element of X must be either a ' +
                                    'graph or an iterable with at least 2 ' +
                                    'and at most 3 elements\n')

                sp_attr_tup.append((S, L))
                ni += 1

            if ni == 0:
                raise ValueError('parsed input is empty')

            return sp_attr_tup

    def pairwise_operation(self, x, y):
        """Calculate shortests paths on attributes.

        Parameters
        ----------
        x, y : tuple
            Tuples of shortest path matrices and their attribute
            dictionaries.

        Returns
        -------
        kernel : number
            The kernel value.

        """
        # Initialise
        Sx, phi_x = x
        Sy, phi_y = y
        kernel = 0
        dimx = Sx.shape[0]
        dimy = Sy.shape[0]
        for i in range(dimx):
            for j in range(dimx):
                if i == j:
                    continue
                for k in range(dimy):
                    for m in range(dimy):
                        if k == m:
                            continue
                        if (Sx[i, j] == Sy[k, m] and
                                Sx[i, j] != float('Inf')):
                            kernel += self.metric(phi_x[i], phi_y[k]) *\
                                self.metric(phi_x[j], phi_y[m])

        return kernel


class ShortestPath(Kernel):
    r"""The shortest path kernel class.

    See :cite:`borgwardt2005shortest`.

    Parameters
    ----------
    algorithm_type : str, default={"dijkstra", "floyd_warshall", "auto"}
        Apply the dijkstra or floyd_warshall algorithm for calculating
        shortest path, or chose automatically ("auto") based on the
        current graph format ("auto").

    with_labels : bool, default=True, case_of_existence=(as_attributes==True)
        Calculate shortest path using graph labels.

    Attributes
    ----------
    X : dict
        A dictionary of pairs between each input graph and a bins where the
        sampled graphlets have fallen.

    _with_labels : bool
        Defines if the shortest path kernel considers also labels.

    _enum : dict
        A dictionary of graph bins holding pynauty objects

    _lt : str
        A label type needed for build shortest path function.

    _lhash : str
        A function for hashing labels, shortest paths.

    _nx : int
        Holds the number of sampled X graphs.

    _ny : int
        Holds the number of sampled Y graphs.

    _X_diag : np.array, shape=(_nx, 1)
        Holds the diagonal of X kernel matrix in a numpy array, if calculated
        (`fit_transform`).

    _phi_X : np.array, shape=(_nx, len(_graph_bins))
        Holds the features of X in a numpy array, if calculated.
        (`fit_transform`).

    Complexity
    ----------
    :math:`O(n*N*|\cup_{i}L_{i}|^{2})`, where :math:`n` the number of graph,
    :math:`N` the number of vertices of the **biggest** Graph and
    :math:`|\cup_{i}L_{i}|` the number of all distinct labels.

    """

    _graph_bins = dict()

    def __init__(self, n_jobs=None,
                 normalize=False,
                 verbose=False,
                 with_labels=True,
                 algorithm_type="auto"):
        """Initialize a `shortest_path` kernel."""
        super(ShortestPath, self).__init__(
            n_jobs=n_jobs, normalize=normalize, verbose=verbose)

        self.with_labels = with_labels
        self.algorithm_type = algorithm_type
        self._initialized.update({"with_labels": False, "algorithm_type": False})

    def initialize(self):
        """Initialize all transformer arguments, needing initialization."""
        if not self._initialized["n_jobs"]:
            if self.n_jobs is not None:
                warnings.warn('no implemented parallelization for ShortestPath')
            self._initialized["n_jobs"] = True

        if not self._initialized["algorithm_type"]:
            if self.algorithm_type == "auto":
                self._graph_format = "auto"
            elif self.algorithm_type == "floyd_warshall":
                self._graph_format = "adjacency"
            elif self.algorithm_type == "dijkstra":
                self._graph_format = "dictionary"
            else:
                raise ValueError('Unsupported "algorithm_type"')

        if not self._initialized["with_labels"]:
            if self.with_labels:
                self._lt = "vertex"
                self._lhash = lhash_labels
                self._decompose_input = decompose_input_labels
            else:
                self._lt = "none"
                self._lhash = lhash
                self._decompose_input = decompose_input

    def transform(self, X):
        """Calculate the kernel matrix, between given and fitted dataset.

        Parameters
        ----------
        X : iterable
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that fitting the given graph
            format). If None the kernel matrix is calculated upon fit data.
            The test samples.

        Returns
        -------
        K : numpy array, shape = [n_targets, n_input_graphs]
            corresponding to the kernel matrix, a calculation between
            all pairs of graphs between target an features

        """
        self._method_calling = 3
        # Check is fit had been called
        check_is_fitted(self, ['X', '_nx', '_enum'])

        # Input validation and parsing
        if X is None:
            raise ValueError('transform input cannot be None')
        else:
            Y = self.parse_input(X)

        # Transform - calculate kernel matrix
        try:
            check_is_fitted(self, ['_phi_X'])
            phi_x = self._phi_X
        except NotFittedError:
            phi_x = np.zeros(shape=(self._nx, len(self._enum)))
            for i in self.X.keys():
                for j in self.X[i].keys():
                    phi_x[i, j] = self.X[i][j]
            self._phi_X = phi_x

        phi_y = np.zeros(shape=(self._ny, len(self._enum) + len(self._Y_enum)))
        for i in Y.keys():
            for j in Y[i].keys():
                phi_y[i, j] = Y[i][j]

        # store _phi_Y for independent (of normalization arg diagonal-calls)
        self._phi_Y = phi_y
        km = np.dot(phi_y[:, :len(self._enum)], phi_x.T)
        self._is_transformed = True
        if self.normalize:
            X_diag, Y_diag = self.diagonal()
            return km / np.sqrt(np.outer(Y_diag, X_diag))
        else:
            return km

    def diagonal(self):
        """Calculate the kernel matrix diagonal for fitted data.

        A funtion called on transform on a seperate dataset to apply
        normalization on the exterior.

        Parameters
        ----------
        None.

        Returns
        -------
        X_diag : np.array
            The diagonal of the kernel matrix, of the fitted data.
            This consists of kernel calculation for each element with itself.

        Y_diag : np.array
            The diagonal of the kernel matrix, of the transformed data.
            This consists of kernel calculation for each element with itself.

        """
        # Check is fit and transform had been called
        try:
            check_is_fitted(self, ['_phi_X'])
        except NotFittedError:
            check_is_fitted(self, ['X'])
            # calculate feature matrices.
            phi_x = np.zeros(shape=(self._nx, len(self._enum)))

            for i in self.X.keys():
                for j in self.X[i].keys():
                    phi_x[i, j] = self.X[i][j]
                    # Transform - calculate kernel matrix
            self._phi_X = phi_x

        try:
            check_is_fitted(self, ['X_diag'])
        except NotFittedError:
            # Calculate diagonal of X
            self._X_diag = np.sum(np.square(self._phi_X), axis=1)
            self._X_diag = np.reshape(self._X_diag, (self._X_diag.shape[0], 1))

        try:
            check_is_fitted(self, ['_phi_Y'])
            # Calculate diagonal of Y
            Y_diag = np.sum(np.square(self._phi_Y), axis=1)
            return self._X_diag, Y_diag
        except NotFittedError:
            return self._X_diag

    def fit_transform(self, X, y=None):
        """Fit and transform, on the same dataset.

        Parameters
        ----------
        X : iterable
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that fitting the given graph
            format).

        y : Object, default=None
            Ignored argument, added for the pipeline.

        Returns
        -------
        K : numpy array, shape = [n_targets, n_input_graphs]
            corresponding to the kernel matrix, a calculation between
            all pairs of graphs between target an features

        """
        self._method_calling = 2
        self.fit(X)

        # calculate feature matrices.
        phi_x = np.zeros(shape=(self._nx, len(self._enum)))

        for i in self.X.keys():
            for j in self.X[i].keys():
                phi_x[i, j] = self.X[i][j]

        # Transform - calculate kernel matrix
        self._phi_X = phi_x
        km = np.dot(phi_x, phi_x.T)

        self._X_diag = np.diagonal(km)
        if self.normalize:
            return np.divide(km, np.sqrt(np.outer(self._X_diag, self._X_diag)))
        else:
            return km

    def parse_input(self, X):
        """Parse and create features for "shortest path" kernel.

        Parameters
        ----------
        X : iterable
            For the input to pass the test, we must have:
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that correspond to the given
            graph format). A valid input also consists of graph type objects.

        Returns
        -------
        sp_counts : dict
            A dictionary that for each vertex holds the counts of shortest path
            tuples.

        """
        if not isinstance(X, collections.Iterable):
            raise TypeError('input must be an iterable\n')
            # Not a dictionary
        else:
            i = -1
            sp_counts = dict()
            if self._method_calling == 1:
                self._enum = dict()
            elif self._method_calling == 3:
                self._Y_enum = dict()
            for (idx, x) in enumerate(iter(X)):
                is_iter = isinstance(x, collections.Iterable)
                if is_iter:
                    x = list(x)
                if is_iter and (len(x) == 0 or
                                (len(x) == 1 and not self.with_labels) or
                                len(x) in [2, 3]):
                    if len(x) == 0:
                        warnings.warn('Ignoring empty element on index: '
                                      + str(idx))
                        continue
                    elif len(x) == 1:
                        spm_data = Graph(x[0], {}, {}, self._graph_format
                                         ).build_shortest_path_matrix(self.algorithm_type,
                                                                      labels=self._lt)
                    else:
                        spm_data = Graph(x[0], x[1], {}, self._graph_format
                                         ).build_shortest_path_matrix(self.algorithm_type,
                                                                      labels=self._lt)
                elif type(x) is Graph:
                    spm_data = x.build_shortest_path_matrix(self.algorithm_type, labels=self._lt)
                else:
                    raise TypeError('each element of X must have at least' +
                                    ' one and at most 3 elements\n')
                i += 1

                S, L = self._decompose_input(spm_data)
                sp_counts[i] = dict()
                for u in range(S.shape[0]):
                    for v in range(S.shape[1]):
                        if u == v or S[u, v] == float("Inf"):
                            continue
                        label = self._lhash(S, u, v, *L)
                        if label not in self._enum:
                            if self._method_calling == 1:
                                idx = len(self._enum)
                                self._enum[label] = idx
                            elif self._method_calling == 3:
                                if label not in self._Y_enum:
                                    idx = len(self._enum) + len(self._Y_enum)
                                    self._Y_enum[label] = idx
                                else:
                                    idx = self._Y_enum[label]
                        else:
                            idx = self._enum[label]
                        if idx in sp_counts[i]:
                            sp_counts[i][idx] += 1
                        else:
                            sp_counts[i][idx] = 1

            if i == -1:
                raise ValueError('parsed input is empty')

            if self._method_calling == 1:
                self._nx = i+1
            elif self._method_calling == 3:
                self._ny = i+1
            return sp_counts


def lhash(S, u, v, *args):
    return S[u, v]


def decompose_input(a):
    return (a, [])


def lhash_labels(S, u, v, *args):
    return (args[0][u], args[0][v], S[u, v])


def decompose_input_labels(args):
    return (args[0], args[1:])



"""Tests for the kernel sub-module."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
import os
import numpy as np

from time import time
from warnings import warn

from numpy.testing import assert_array_less

from sklearn.model_selection import train_test_split

from grakel.datasets import fetch_dataset
from grakel.datasets import get_dataset_info
from grakel.datasets.base import read_data

from grakel.kernels import GraphletSampling
from grakel.kernels import RandomWalk
from grakel.kernels import RandomWalkLabeled
from grakel.kernels import ShortestPath
from grakel.kernels import ShortestPathAttr
from grakel.kernels import WeisfeilerLehman
from grakel.kernels import WeisfeilerLehmanOptimalAssignment
from grakel.kernels import NeighborhoodHash
from grakel.kernels import PyramidMatch
from grakel.kernels import SubgraphMatching
from grakel.kernels import NeighborhoodSubgraphPairwiseDistance
from grakel.kernels import LovaszTheta
from grakel.kernels import SvmTheta
from grakel.kernels import OddSth
from grakel.kernels import Propagation
from grakel.kernels import PropagationAttr
from grakel.kernels import HadamardCode
from grakel.kernels import MultiscaleLaplacian
from grakel.kernels import VertexHistogram
from grakel.kernels import EdgeHistogram
from grakel.kernels import GraphHopper
from grakel.kernels import CoreFramework

global verbose, main, development

fdir = os.path.dirname(__file__)

default_eigvalue_precision = float("-1e-5")

# Mark optional dependencies
cvxopt = True
try:
    import cvxopt
except ImportError:
    cvxopt = False


if __name__ == '__main__':
    import argparse
    # Create an argument parser for the installer of pynauty
    parser = argparse.ArgumentParser(description='A test file for all kernels')

    parser.add_argument(
        '--verbose',
        help='print kernels with their outputs on stdout',
        action="store_true")
    parser.add_argument(
        '--problematic',
        help='allow execution of problematic test cases in development',
        action="store_true")
    parser.add_argument(
        '--slow',
        help='allow execution of slow test cases in development',
        action="store_true")
    parser.add_argument(
        '--ignore_warnings',
        help='ignore warnings produced by kernel executions',
        action="store_true")

    parser.add_argument(
        '--time',
        help='time the kernel computation (has effect only on verbose)',
        action="store_true")
    parser.add_argument(
        '--dataset',
        help='choose the dataset for tests requiring node/edge labels',
        type=str,
        default="MUTAG"
    )

    parser.add_argument(
        '--dataset_attr',
        help='choose the dataset for tests requiring node attributes',
        type=str,
        default="Cuneiform"
    )

    parser.add_argument('--normalize', help='normalize the kernel output',
                        action="store_true")

    meg = parser.add_mutually_exclusive_group()
    meg.add_argument(
        '--develop',
        help='execute only tests connected with current development',
        action="store_true")
    meg.add_argument(
        '--all',
        help='execute all tests',
        action="store_true")
    meg.add_argument(
        '--main',
        help='execute the main tests [default]',
        action="store_true")

    args = parser.parse_args()

    verbose = bool(args.verbose)
    if args.all:
        main, develop = True, True
    elif args.develop:
        main, develop = False, True
    else:
        main, develop = True, False
    problematic = bool(args.problematic)
    slow = bool(args.slow)

    if bool(args.ignore_warnings):
        import warnings
        warnings.filterwarnings('ignore', category=UserWarning)

    normalize = bool(args.normalize)
    time_kernel = bool(args.time)
    dataset_name = args.dataset
    dataset_attr_name = args.dataset_attr

else:
    import warnings
    warnings.filterwarnings('ignore', category=UserWarning)
    main, develop, problematic, slow = True, False, False, False
    normalize, verbose, time_kernel = False, False, False
    dataset_name = "MUTAG"
    dataset_attr_name = "Cuneiform"

# consistency check for the dataset
dinfo = get_dataset_info(dataset_name)
if dinfo is None:
    raise TypeError('dataset not found')
elif not dinfo["nl"] and not dinfo["el"]:
    raise TypeError('dataset must have either node and edge labels')

# consistency check for the attribute dataset
dinfo_attr = get_dataset_info(dataset_attr_name)
if dinfo is None:
    raise TypeError('dataset for attributes not found')
elif not dinfo_attr["nl"] and not dinfo_attr["el"]:
    raise TypeError('dataset must have node attributes')


# The baseline dataset for node, edge_labels
global dataset, dataset_tr, dataset_te

try:
    dataset = fetch_dataset(dataset_name, with_classes=False, verbose=verbose).data
except Exception:
    # Offline testing
    warn('There was a problem fetching dataset for attributes: [' + dataset_name + ']')
    if dataset_name != 'MUTAG':
        warn('Switching back to baseline dataset MUTAG')
    warn('Using an offline version..')
    cwd = os.getcwd()
    os.chdir(os.path.join(fdir, 'data'))
    dataset = read_data('MUTAG', with_classes=False).data
    os.chdir(cwd)

dataset_tr, dataset_te = train_test_split(dataset,
                                          test_size=0.2,
                                          random_state=42)

# The baseline dataset for node/edge-attributes
global dataset_attr, dataset_attr_tr, dataset_attr_te

try:
    dataset_attr = fetch_dataset(dataset_attr_name, with_classes=False,
                                 prefer_attr_nodes=True,
                                 prefer_attr_edges=True,
                                 verbose=verbose).data
except Exception:
    # Offline testing
    warn('There was a problem fetching dataset for attributes: [' + dataset_attr_name + ']')
    if dataset_attr_name != 'Cuneiform':
        warn('Switching back to baseline dataset Cuneiform')
    warn('Using an offline version..')
    cwd = os.getcwd()
    os.chdir(os.path.join(fdir, 'data'))
    dataset_attr = read_data('Cuneiform', with_classes=False, prefer_attr_nodes=True,
                             prefer_attr_edges=True).data
    os.chdir(cwd)

dataset_attr_tr, dataset_attr_te = train_test_split(dataset_attr,
                                                    test_size=0.2,
                                                    random_state=42)


def test_random_walk():
    """Eigenvalue test for the Simple, Labelled Random Walk kernel."""
    rw_kernel = RandomWalk(verbose=verbose, normalize=normalize, lamda=0.01)
    if verbose:
        print_kernel("Random Walk", rw_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(rw_kernel, dataset)

    rw_kernel_lab = RandomWalkLabeled(verbose=verbose, normalize=normalize, lamda=0.0001)
    if verbose:
        print_kernel("Random Walk Labelled", rw_kernel_lab, dataset_tr, dataset_te)
    else:
        positive_eig(rw_kernel_lab, dataset)


def test_shortest_path():
    """Eigenvalue test for the Shortest Path kernel."""
    sp_kernel = ShortestPath(verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("Shortest Path", sp_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(sp_kernel, dataset)

    sp_kernel_attr = ShortestPathAttr(verbose=verbose, normalize=normalize)
    if verbose and slow:
        print_kernel("Shortest Path Attr", sp_kernel_attr, dataset_attr_tr, dataset_attr_te)
    elif not verbose and slow:
        positive_eig(sp_kernel_attr, dataset_attr)


def test_graphlet_sampling():
    """Eigenvalue test for the Graphlet Sampling Kernel."""
    gs_kernel = GraphletSampling(verbose=verbose, normalize=normalize, sampling=dict(n_samples=150))
    if verbose:
        print_kernel("Graphlet Sampling", gs_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(gs_kernel, dataset)


def test_weisfeiler_lehman():
    """Eigenvalue test for the Weisfeiler Lehman kernel."""
    wl_st_kernel = WeisfeilerLehman(verbose=verbose, normalize=normalize,
                                    base_graph_kernel=VertexHistogram)
    if verbose:
        print_kernel("WL/Subtree", wl_st_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(wl_st_kernel, dataset)


def test_weisfeiler_lehman_optimal_assignment():
    """Eigenvalue test for the Weisfeiler Lehman Optimal Assignment kernel."""
    wl_oa_kernel = WeisfeilerLehmanOptimalAssignment(
        verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("WL-OA", wl_oa_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(wl_oa_kernel, dataset)


def test_pyramid_match():
    """Eigenvalue test for the Pyramid Match kernel."""
    pm_kernel = PyramidMatch(verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("Pyramid Match", pm_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(pm_kernel, dataset)


def test_neighborhood_hash():
    """Eigenvalue test for the Neighborhood Hash kernel."""
    nh_kernel = NeighborhoodHash(verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("Neighborhood Hash", nh_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(nh_kernel, dataset)


def test_subgraph_matching():
    """Eigenvalue test for the subgraph_matching kernel."""
    sm_kernel = SubgraphMatching(verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("Subgraph Matching", sm_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(sm_kernel, dataset)


def test_neighborhood_subgraph_pairwise_distance():
    """Eigenvalue test for the neighborhood subgraph pairwise distance kernel."""
    nspd_kernel = NeighborhoodSubgraphPairwiseDistance(
        verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("NSPD", nspd_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(nspd_kernel, dataset)


if cvxopt:
    def test_lovasz_theta():
        """Eigenvalue test for the Lovasz-theta distance kernel."""
        lt_kernel = LovaszTheta(verbose=verbose, normalize=normalize)

        if verbose:
            print_kernel("Lovasz-theta", lt_kernel, dataset_tr, dataset_te)
        else:
            positive_eig(lt_kernel, dataset)


def test_svm_theta():
    """Eigenvalue test for the SVM-theta distance kernel."""
    svm_kernel = SvmTheta(verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("SVM-theta", svm_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(svm_kernel, dataset)


def test_odd_sth():
    """Eigenvalue test for the ODD-STh kernel."""
    odd_sth_kernel = OddSth(verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("ODD-STh", odd_sth_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(odd_sth_kernel, dataset)


def test_propagation():
    """Eigenvalue test for the Propagation kernel."""
    propagation_kernel = Propagation(verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("Propagation", propagation_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(propagation_kernel, dataset)

    propagation_kernel_attr = PropagationAttr(verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("Propagation", propagation_kernel_attr, dataset_attr_tr, dataset_attr_te)
    else:
        positive_eig(propagation_kernel_attr, dataset_attr)


def test_hadamard_code():
    """Eigenvalue test for the Hadamard Code kernel."""
    hadamard_code_kernel = HadamardCode(verbose=verbose, normalize=normalize,
                                        base_graph_kernel=VertexHistogram)
    if verbose:
        print_kernel("Hadamard-Code/VH [Simple]",
                     hadamard_code_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(hadamard_code_kernel, dataset)


def test_multiscale_laplacian():
    """Eigenvalue test for the Multiscale Laplacian kernel."""
    mlf_kernel = MultiscaleLaplacian(verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("Multiscale Laplacian", mlf_kernel,
                     dataset_attr_tr, dataset_attr_te)
    else:
        positive_eig(mlf_kernel, dataset_attr)


def test_vertex_histogram():
    """Eigenvalue test for the Vertex Histogram Kernel."""
    vh_kernel = VertexHistogram(verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("Vertex Histogram", vh_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(vh_kernel, dataset)


def test_edge_histogram():
    """Eigenvalue test for the Edge Histogram Kernel."""
    eh_kernel = EdgeHistogram(verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("Edge Histogram", eh_kernel, dataset_tr, dataset_te)
    else:
        positive_eig(eh_kernel, dataset)


def test_graph_hopper():
    """Eigenvalue test for the Graph Hopper Kernel."""
    gh_kernel = GraphHopper(verbose=verbose, normalize=normalize)
    if verbose:
        print_kernel("Graph Hopper", gh_kernel, dataset_attr_tr, dataset_attr_te)
    else:
        positive_eig(gh_kernel, dataset_attr)


def test_core_framework():
    """Eigenvalue test for the Core kernel Framework."""
    base_graph_kernel = (WeisfeilerLehman, dict(base_graph_kernel=VertexHistogram))
    core_framework = CoreFramework(verbose=verbose, normalize=normalize, base_graph_kernel=base_graph_kernel)
    if verbose:
        print_kernel("Core Framework", core_framework, dataset_tr, dataset_te)
    else:
        positive_eig(core_framework, dataset)


def print_kernel(name, kernel, X, Y):
    """Print kernels in case of verbose execution."""
    if time_kernel:
        print(str(name) + ":\n" + (len(str(name)) * "-") + "-\n")
        print("fit_transform\n-------------")

        # [time] fit_transform
        start = time()
        Kft = kernel.fit_transform(X)
        ft_time = time() - start

        print(Kft)
        print("[TIME] fit_transform:", sec_to_time(ft_time))
        print("\ntransform\n---------")

        start = time()
        Kt = kernel.transform(Y)
        t_time = time() - start
        print(Kt)
        print("[TIME] transform:", sec_to_time(t_time))
        print("[TIME] total:", sec_to_time(ft_time+t_time))
        print("--------------------------------------" +
              "--------------------------------------\n")
    else:
        print(str(name) + ":\n" + (len(str(name)) * "-") + "-\n")
        print("fit_transform\n-------------")
        print(kernel.fit_transform(X))

        print("\ntransform\n---------")
        print(kernel.transform(Y))
        print("--------------------------------------" +
              "--------------------------------------\n")


def sec_to_time(sec):
    """Print time in a correct format."""
    dt = list()
    days = int(sec // 86400)
    if days > 0:
        sec -= 86400*days
        dt.append(str(days) + " d")

    hrs = int(sec // 3600)
    if hrs > 0:
        sec -= 3600*hrs
        dt.append(str(hrs) + " h")

    mins = int(sec // 60)
    if mins > 0:
        sec -= 60*mins
        dt.append(str(mins) + " m")

    if sec > 0:
        dt.append(str(round(sec, 2)) + " s")
    return " ".join(dt)


def positive_eig(kernel, X):
    """Assert true if the calculated kernel matrix is valid."""
    K = kernel.fit_transform(X)
    min_eig = np.real(np.min(np.linalg.eig(K)[0]))
    assert_array_less(default_eigvalue_precision, min_eig)


if verbose and main:
    test_random_walk()
    test_shortest_path()
    test_weisfeiler_lehman()
    test_weisfeiler_lehman_optimal_assignment()
    test_neighborhood_hash()
    test_graphlet_sampling()
    test_lovasz_theta()
    test_svm_theta()
    test_odd_sth()
    test_propagation()
    test_hadamard_code()
    test_neighborhood_subgraph_pairwise_distance()
    test_pyramid_match()
    test_multiscale_laplacian()
    test_edge_histogram()
    test_vertex_histogram()
    test_subgraph_matching()
    test_graph_hopper()
    test_core_framework()



"""The main function for the tests sub-module."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause

if __name__ == '__main__':
    import os
    import sys
    import warnings

    from subprocess import check_call

    warnings.filterwarnings('ignore', category=UserWarning)

    python_executable_address = str(sys.executable)
    test_dir = str(os.path.dirname(os.path.realpath(__file__)))
    project_dir = str(os.path.realpath(os.path.join(__file__, "../../../")))

    print('Installing the latest "GraKeL"..')
    print('--------------------------------')

    cwd = os.getcwd()
    os.chdir(project_dir)
    try:
        check_call([python_executable_address, project_dir + "/setup.py",
                    "install"])
    finally:
        os.chdir(cwd)

    print('................................................................\n')

    print('Testing Graph..')
    print('---------------')
    check_call([python_executable_address, test_dir + "/test_graph.py",
                "--ignore_warnings", "--verbose"])
    print('................................................................\n')

    print('Testing Kernels..')
    print('-----------------')
    check_call([python_executable_address, test_dir + "/test_kernels.py",
                "--verbose", "--time", "--ignore_warnings", "--all"])
    print('................................................................\n')

    print('Testing Graph Kernels..')
    print('-----------------------')
    check_call([python_executable_address,
                test_dir + "/test_graph_kernels.py",
                "--verbose", "--time", "--ignore_warnings", "--all"])
    print('................................................................')



"""Import datasets related with graph kernels, from a large collection."""
from grakel.datasets.base import fetch_dataset
from grakel.datasets.base import get_dataset_info
from grakel.datasets.testing import generate_dataset

__all__ = [
    "get_dataset_info",
    "fetch_dataset",
    "generate_dataset"
]



"""__init__ file for kernel sub-module of grakel."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
from grakel.kernels.kernel import Kernel

from grakel.kernels.graphlet_sampling import GraphletSampling
from grakel.kernels.random_walk import RandomWalk
from grakel.kernels.random_walk import RandomWalkLabeled
from grakel.kernels.shortest_path import ShortestPath
from grakel.kernels.shortest_path import ShortestPathAttr
from grakel.kernels.weisfeiler_lehman import WeisfeilerLehman
from grakel.kernels.neighborhood_hash import NeighborhoodHash
from grakel.kernels.pyramid_match import PyramidMatch
from grakel.kernels.subgraph_matching import SubgraphMatching
from grakel.kernels.neighborhood_subgraph_pairwise_distance import \
    NeighborhoodSubgraphPairwiseDistance
from grakel.kernels.lovasz_theta import LovaszTheta
from grakel.kernels.svm_theta import SvmTheta
from grakel.kernels.odd_sth import OddSth
from grakel.kernels.propagation import Propagation
from grakel.kernels.propagation import PropagationAttr
from grakel.kernels.hadamard_code import HadamardCode
from grakel.kernels.multiscale_laplacian import MultiscaleLaplacian
from grakel.kernels.vertex_histogram import VertexHistogram
from grakel.kernels.edge_histogram import EdgeHistogram
from grakel.kernels.graph_hopper import GraphHopper
from grakel.kernels.core_framework import CoreFramework
from grakel.kernels.weisfeiler_lehman_optimal_assignment import WeisfeilerLehmanOptimalAssignment

__all__ = [
    "default_executor",
    "Kernel",
    "GraphletSampling",
    "RandomWalk",
    "RandomWalkLabeled",
    "ShortestPath",
    "ShortestPathAttr",
    "WeisfeilerLehman",
    "NeighborhoodHash",
    "PyramidMatch",
    "SubgraphMatching",
    "NeighborhoodSubgraphPairwiseDistance",
    "LovaszTheta",
    "SvmTheta",
    "OddSth",
    "Propagation",
    "PropagationAttr",
    "HadamardCode",
    "MultiscaleLaplacian",
    "VertexHistogram",
    "EdgeHistogram",
    "GraphHopper",
    "CoreFramework",
    "WeisfeilerLehmanOptimalAssignment"
]



"""Tests for the GraphKernel class."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
import os
from time import time
from warnings import warn

from sklearn.model_selection import train_test_split

from grakel.datasets import fetch_dataset
from grakel.datasets.base import read_data
from grakel.datasets import get_dataset_info
from grakel.graph_kernels import GraphKernel


fdir = os.path.dirname(__file__)

global verbose, main, development

if __name__ == '__main__':
    import argparse

    # Create an argument parser for the installer of pynauty
    parser = argparse.ArgumentParser(description='A test file for all kernels')

    parser.add_argument('--verbose', help='print kernels with their outputs' +
                        ' on stdout', action="store_true")
    parser.add_argument('--time', help='time the kernel computation (has ef' +
                        'fect only on verbose)', action="store_true")
    parser.add_argument('--problematic', help='allow execution of problemati' +
                        'c test cases in development', action="store_true")
    parser.add_argument('--slow', help='allow execution of slow test cases' +
                        ' in development', action="store_true")
    parser.add_argument('--normalize', help='normalize the kernel output',
                        action="store_true")
    parser.add_argument('--ignore_warnings', help='ignore warnings produced ' +
                        'by kernel executions', action="store_true")

    parser.add_argument(
        '--dataset',
        help='choose the dataset for tests requiring node/edge labels',
        type=str,
        default="MUTAG"
    )

    parser.add_argument(
        '--dataset_attr',
        help='choose the dataset for tests requiring node attributes',
        type=str,
        default="Cuneiform"
    )

    meg = parser.add_mutually_exclusive_group()
    meg.add_argument('--develop', help='execute only tests connected with ' +
                     'current development', action="store_true")
    meg.add_argument('--all', help='execute all tests', action="store_true")
    meg.add_argument('--main', help='execute the main tests [default]',
                     action="store_true")

    args = parser.parse_args()

    verbose = bool(args.verbose)
    time_kernel = bool(args.time)
    if args.all:
        main, develop = True, True
    elif args.develop:
        main, develop = False, True
    else:
        main, develop = True, False
    problematic = bool(args.problematic)
    slow = bool(args.slow)

    if bool(args.ignore_warnings):
        import warnings
        warnings.filterwarnings('ignore', category=UserWarning)

    normalize = bool(args.normalize)

    dataset_name = args.dataset
    dataset_attr_name = args.dataset_attr

else:
    import warnings
    warnings.filterwarnings('ignore', category=UserWarning)
    main, develop, problematic, slow = True, False, False, False
    normalize, verbose, time_kernel = False, False, False
    dataset_name = "MUTAG"
    dataset_attr_name = "Cuneiform"

global dataset_tr, dataset_te, dataset_attr_tr, dataset_attr_te

# consistency check for the dataset
dinfo = get_dataset_info(dataset_name)
if dinfo is None:
    raise TypeError('dataset not found')
elif not dinfo["nl"] and not dinfo["el"]:
    raise TypeError('dataset must have either node and edge labels')

# consistency check for the attribute dataset
dinfo_attr = get_dataset_info(dataset_attr_name)
if dinfo is None:
    raise TypeError('dataset for attributes not found')
elif not dinfo_attr["nl"] and not dinfo_attr["el"]:
    raise TypeError('dataset must have node attributes')


# The baseline dataset for node, edge_labels
global dataset, dataset_tr, dataset_te

try:
    dataset = fetch_dataset(dataset_name, with_classes=False, verbose=verbose).data
except Exception:
    # Offline testing
    warn('There was a problem fetching dataset for attributes: [' + dataset_name + ']')
    if dataset_name != 'MUTAG':
        warn('Switching back to baseline dataset MUTAG')
    warn('Using an offline version..')
    cwd = os.getcwd()
    os.chdir(os.path.join(fdir, 'data'))
    dataset = read_data('MUTAG', with_classes=False).data
    os.chdir(cwd)

dataset_tr, dataset_te = train_test_split(dataset,
                                          test_size=0.2,
                                          random_state=42)

# The baseline dataset for node/edge-attributes
global dataset_attr, dataset_attr_tr, dataset_attr_te

try:
    dataset_attr = fetch_dataset(dataset_attr_name, with_classes=False,
                                 prefer_attr_nodes=True,
                                 prefer_attr_edges=True,
                                 verbose=verbose).data
except Exception:
    # Offline testing
    warn('There was a problem fetching dataset for attributes: [' + dataset_attr_name + ']')
    if dataset_attr_name != 'Cuneiform':
        warn('Switching back to baseline dataset Cuneiform')
    warn('Using an offline version..')
    cwd = os.getcwd()
    os.chdir(os.path.join(fdir, 'data'))
    dataset_attr = read_data('Cuneiform', with_classes=False, prefer_attr_nodes=True,
                             prefer_attr_edges=True).data
    os.chdir(cwd)

dataset_attr_tr, dataset_attr_te = train_test_split(dataset_attr,
                                                    test_size=0.2,
                                                    random_state=42)

if verbose:
    # Guard for nosetests
    def test_random_walk():
        """Test the Simple random walk kernel [decorator]."""
        gk = GraphKernel(kernel={"name": "random_walk"}, verbose=verbose,
                         normalize=normalize)
        if verbose:
            print_kernel_decorator("Random Walk", gk, dataset_tr, dataset_te)

    def test_shortest_path():
        """Test the Shortest Path kernel [decorator]."""
        gk = GraphKernel(kernel={"name": "shortest_path"}, verbose=verbose,
                         normalize=normalize)
        if verbose:
            print_kernel_decorator("Shortest Path", gk, dataset_tr, dataset_te)

    def test_graphlet_sampling():
        """Test the Graphlet Sampling Kernel [decorator]."""
        gk = GraphKernel(kernel={"name": "graphlet_sampling",
                                 "sampling": {"n_samples": 200}},
                         verbose=verbose, normalize=normalize)

        if verbose:
            print_kernel_decorator("Graphlet Sampling", gk, dataset_tr, dataset_te)

    def test_weisfeiler_lehman():
        """Test the Weisfeiler Lehman kernel [decorator]."""
        gk = GraphKernel(kernel=[{"name": "weisfeiler_lehman"},
                         {"name": "vertex_histogram"}],
                         verbose=verbose, normalize=normalize)
        if verbose:
            print_kernel_decorator("WL/Subtree", gk, dataset_tr, dataset_te)

    def test_weisfeiler_lehman_optimal_assignment():
        """Test the Weisfeiler Lehman Optimal Assignment kernel [decorator]."""
        gk = GraphKernel(kernel="WL-OA", verbose=verbose, normalize=normalize)
        if verbose:
            print_kernel_decorator("WL-OA", gk, dataset_tr, dataset_te)

    def test_pyramid_match():
        """Test the Pyramid Match kernel [decorator]."""
        gk = GraphKernel(kernel={"name": "pyramid_match"}, verbose=verbose,
                         normalize=normalize)
        if verbose:
            print_kernel_decorator("Pyramid Match", gk, dataset_tr, dataset_te)

    def test_neighborhood_hash():
        """Test the Neighborhood Hash kernel [decorator]."""
        gk = GraphKernel(kernel={"name": "neighborhood_hash"}, verbose=verbose,
                         normalize=normalize)
        if verbose:
            print_kernel_decorator("Neighborhood Hash", gk, dataset_tr, dataset_te)

    def test_subgraph_matching():
        """Test the Subgraph Matching kernel [decorator]."""
        gk = GraphKernel(kernel={"name": "subgraph_matching"}, verbose=verbose,
                         normalize=normalize)
        if verbose:
            print_kernel_decorator("Subgraph Matching", gk, dataset_tr, dataset_te)

    def test_neighborhood_pairwise_distance():
        """Test the Neighborhood Subgraph Pairwise Distance kernel [decorator]."""
        gk = GraphKernel(kernel={
            "name": "neighborhood_subgraph_pairwise_distance"},
                         verbose=verbose, normalize=normalize)

        if verbose:
            print_kernel_decorator("NSPD", gk, dataset_tr, dataset_te)

    def test_lovasz_theta():
        """Test the Lovasz-theta kernel [decorator]."""
        try:
            gk = GraphKernel(kernel={"name": "lovasz_theta"},
                             verbose=verbose, normalize=normalize)
        except ImportError:
            return

        if verbose:
            print_kernel_decorator("Lovasz-theta", gk, dataset_tr, dataset_te)

    def test_svm_theta():
        """Test the SVM-theta kernel [decorator]."""
        gk = GraphKernel(kernel={"name": "svm_theta"},
                         verbose=verbose, normalize=normalize)

        if verbose:
            print_kernel_decorator("SVM-theta", gk, dataset_tr, dataset_te)

    def test_odd_sth():
        """Test the ODD-STh kernel [decorator]."""
        gk = GraphKernel(kernel={"name": "odd_sth"},
                         verbose=verbose, normalize=normalize)

        if verbose:
            print_kernel_decorator("ODD-STh", gk, dataset_tr, dataset_te)

    def test_propagation():
        """Test the Propagation kernel [decorator]."""
        gk = GraphKernel(kernel={"name": "propagation"},
                         verbose=verbose, normalize=normalize)

        if verbose:
            print_kernel_decorator("Propagation", gk, dataset_tr, dataset_te)

    def test_hadamard_code():
        """Test the Hadamard Code kernel [decorator]."""
        gk = GraphKernel(kernel=[{"name": "hadamard_code"},
                                 {"name": "subtree_wl"}],
                         verbose=verbose, normalize=normalize)

        if verbose:
            print_kernel_decorator("Hadamard-Code/Subtree-WL [Simple]",
                                   gk, dataset_tr, dataset_te)

    def test_edge_histogram():
        """Test the Edge Histogram kernel [decorator]."""
        gk = GraphKernel(kernel={"name": "edge_histogram"},
                         verbose=verbose, normalize=normalize)

        if verbose:
            print_kernel_decorator("Edge Histogram",
                                   gk, dataset_tr, dataset_te)

    def test_vertex_histogram():
        """Test the Vertex Histogram kernel [decorator]."""
        gk = GraphKernel(kernel={"name": "vertex_histogram"},
                         verbose=verbose, normalize=normalize)

        if verbose:
            print_kernel_decorator("Vertex Histogram",
                                   gk, dataset_tr, dataset_te)

    def test_multiscale_laplacian():
        """Test the Multiscale Laplacian kernel [decorator]."""
        gk = GraphKernel(kernel="ML", verbose=verbose, normalize=normalize)

        if verbose:
            print_kernel_decorator("Multiscale Laplacian",
                                   gk, dataset_attr_tr, dataset_attr_te)

    def test_graph_hopper():
        """Test the Graph Hopper kernel [decorator]."""
        gk = GraphKernel(kernel={"name": "graph_hopper"},
                         verbose=verbose, normalize=normalize)

        if verbose:
            print_kernel_decorator("Graph Hopper",
                                   gk, dataset_attr_tr, dataset_attr_te)

    def test_core_framework():
        """Test the Core Framework kernel [decorator]."""
        kernel = [{"name": "core_framework"}, {"name": "weisfeiler_lehman"},
                  {"name": "vertex_histogram"}]
        gk = GraphKernel(kernel=kernel, verbose=verbose, normalize=normalize)
        if verbose:
            print_kernel_decorator("Core Framework",
                                   gk, dataset_tr, dataset_te)

    def sec_to_time(sec):
        """Print time in a correct format."""
        dt = list()
        days = int(sec // 86400)
        if days > 0:
            sec -= 86400*days
            dt.append(str(days) + " d")

        hrs = int(sec // 3600)
        if hrs > 0:
            sec -= 3600*hrs
            dt.append(str(hrs) + " h")

        mins = int(sec // 60)
        if mins > 0:
            sec -= 60*mins
            dt.append(str(mins) + " m")

        if sec > 0:
            dt.append(str(round(sec, 2)) + " s")
        return " ".join(dt)

    def print_kernel_decorator(name, kernel, X, Y):
        """Print kernels in case of verbose execution."""
        if time_kernel:
            name += " [decorator]"
            print(str(name) + ":\n" + (len(str(name)) * "-") + "-\n")
            print("fit_transform\n-------------")

            # [time] fit_transform
            start = time()
            Kft = kernel.fit_transform(X)
            ft_time = time() - start

            print(Kft)
            print("[TIME] fit_transform:", sec_to_time(ft_time))
            print("\ntransform\n---------")

            start = time()
            Kt = kernel.transform(Y)
            t_time = time() - start
            print(Kt)
            print("[TIME] transform:", sec_to_time(t_time))
            print("[TIME] total:", sec_to_time(ft_time+t_time))
            print("--------------------------------------" +
                  "--------------------------------------\n")
        else:
            name += " [decorator]"
            print(str(name) + ":\n" + (len(str(name)) * "-") + "-\n")
            print("fit_transform\n-------------")
            print(kernel.fit_transform(X))

            print("\ntransform\n---------")
            print(kernel.transform(Y))
            print("--------------------------------------" +
                  "--------------------------------------\n")

if verbose and main:
    test_random_walk()
    test_shortest_path()
    test_weisfeiler_lehman()
    test_weisfeiler_lehman_optimal_assignment()
    test_pyramid_match()
    test_neighborhood_hash()
    test_graphlet_sampling()
    test_lovasz_theta()
    test_svm_theta()
    test_odd_sth()
    test_propagation()
    test_hadamard_code()
    test_neighborhood_pairwise_distance()
    test_multiscale_laplacian()
    test_subgraph_matching()
    test_vertex_histogram()
    test_edge_histogram()
    test_core_framework()
    test_graph_hopper()



"""The lovasz theta kernel as in :cite:`johansson2014global`."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
import collections
import warnings

import numpy as np

from grakel.kernels import Kernel
from grakel.graph import Graph
from grakel.tools import distribute_samples

from sklearn.utils import check_random_state

from math import sqrt
from numpy import pad
from numpy.linalg import LinAlgError
from numpy.linalg import norm
from scipy.linalg import cholesky
from scipy.linalg import eigvalsh
from scipy.linalg import solve

cvxopt_installed = True
try:
    from cvxopt.base import matrix
    from cvxopt.base import spmatrix
    from cvxopt.solvers import sdp
    from cvxopt.solvers import options
    options['show_progress'] = False
except ImportError:
    cvxopt_installed = False

min_weight = float("1e-10")
angle_precision = float("1e-6")
tolerance = float("1e-1")


class LovaszTheta(Kernel):
    """Lovasz theta kernel as proposed in :cite:`johansson2014global`.

    Parameters
    ----------
    X,Y : *valid-graph-format*
        The pair of graphs on which the kernel is applied.

    n_samples : int, default=50
        The number of samples.

    subsets_size_range : tuple, len=2, default=(2,8)
        (min, max) size of the vertex set of sampled subgraphs.

    random_state :  RandomState or int, default=None
        A random number generator instance or an int to initialize a RandomState as a seed.

    base_kernel : function (np.1darray, np.1darray -> number), default=None

        The applied metric between the lovasz_theta numbers of subgraphs.
        If None :math:`f(x,y) = x*y`

    max_dim : int, default=None
        The maximum graph size that can appear both in fit or transform.
        When None, max_dim is calculated based on the size of the biggest graph on fit.
        This can lead to a crash in case a graph appears in transform with size bigger than in fit.

    Attributes
    ----------
    d_ : int,
        The maximum matrix dimension of fit plus 1. Signifies the number
        of features assigned for lovasz labelling.

    random_state_ : RandomState
        A RandomState object handling all randomness of the class.

    """

    _graph_format = "adjacency"

    def __init__(self,
                 n_jobs=None,
                 normalize=False,
                 verbose=False,
                 random_state=None,
                 n_samples=50,
                 subsets_size_range=(2, 8),
                 max_dim=None,
                 base_kernel=None):
        """Initialise a lovasz_theta kernel."""
        # setup valid parameters and initialise from parent
        if not cvxopt_installed:
            raise ImportError('cvxopt should be installed for the '
                              'computation of the Lovasz-Theta Kernel.')

        super(LovaszTheta, self).__init__(n_jobs=n_jobs,
                                          normalize=normalize,
                                          verbose=verbose)

        self.n_samples = n_samples
        self.subsets_size_range = subsets_size_range
        self.base_kernel = base_kernel
        self.random_state = random_state
        self.max_dim = max_dim
        self._initialized.update({"n_samples": False, "subsets_size_range": False,
                                  "base_kernel": False, "random_state": False,
                                  "max_dim": False})

    def initialize(self):
        """Initialize all transformer arguments, needing initialization."""
        super(LovaszTheta, self).initialize()

        if not self._initialized["n_samples"]:
            if self.n_samples <= 0 or type(self.n_samples) is not int:
                raise TypeError('n_samples must an integer be bigger than '
                                'zero')
            self._initialized["n_samples"] = True

        if not self._initialized["subsets_size_range"]:
            if (type(self.subsets_size_range) is not tuple
                    or len(self.subsets_size_range) != 2
                    or any(type(i) is not int for i in self.subsets_size_range)
                    or self.subsets_size_range[0] > self.subsets_size_range[1]
                    or self.subsets_size_range[0] <= 0):
                raise TypeError('subsets_size_range subset size range'
                                'must be a tuple of two integers in '
                                'increasing order, bigger than 1')
            self._initialized["subsets_size_range"] = True

        if not self._initialized["base_kernel"]:
            if not callable(self.base_kernel) and self.base_kernel is not None:
                raise TypeError('base_kernel between arguments ' +
                                'must be a function')
            self._initialized["base_kernel"] = True
            self.base_kernel_ = inner_product

        if not self._initialized["random_state"]:
            self.random_state_ = check_random_state(self.random_state)
            self._initialized["random_state"] = True

        if not self._initialized["max_dim"]:
            if self.max_dim is not None and (type(self.max_dim) is not int or self.max_dim < 1):
                raise ValueError('max_dim if not None, should be an integer bigger than 1')
            if self.max_dim is None:
                self.d_ = None
            else:
                self.d_ = self.max_dim + 1
            self._initialized["max_dim"] = True

    def parse_input(self, X):
        """Parse and create features for lovasz_theta kernel.

        Parameters
        ----------
        X : iterable
            For the input to pass the test, we must have:
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that correspond to the given
            graph format). A valid input also consists of graph type objects.


        Returns
        -------
        out : list
            The lovasz metrics for the given input.

        """
        if not isinstance(X, collections.Iterable):
            raise TypeError('input must be an iterable\n')
        else:
            i = 0
            adjm = list()
            max_dim = 0
            for (idx, x) in enumerate(iter(X)):
                is_iter = False
                if isinstance(x, collections.Iterable):
                    x, is_iter = list(x), True
                if is_iter and len(x) in [0, 1, 2, 3]:
                    if len(x) == 0:
                        warnings.warn('Ignoring empty element ' +
                                      'on index: '+str(idx))
                        continue
                    else:
                        x = Graph(x[0], {}, {}, self._graph_format)
                elif type(x) is not Graph:
                    raise TypeError('each element of X must be either a ' +
                                    'graph or an iterable with at least 1 ' +
                                    'and at most 3 elements\n')
                i += 1
                A = x.get_adjacency_matrix()
                adjm.append(A)
                max_dim = max(max_dim, A.shape[0])

            if self._method_calling == 1:
                if self.d_ is None:
                    self.d_ = max_dim + 1

            if self.d_ < max_dim + 1:
                if self.max_dim is None and self._method_calling == 3:
                    raise ValueError('Maximum dimension of a graph in transform is bigger '
                                     'than the one found in fit. To avoid that use max_dim parameter.')
                else:
                    raise ValueError('max_dim should correspond to the '
                                     'biggest graph inside the dataset')

            out = list()
            for A in adjm:
                X, t = _calculate_lovasz_embeddings_(A)
                U = _calculate_lovasz_labelling_(X, t, self.d_)
                out.append(self._calculate_MEC_(U))

            if i == 0:
                raise ValueError('parsed input is empty')

            return out

    def pairwise_operation(self, x, y):
        """Lovasz theta kernel as proposed in :cite:`johansson2014global`.

        Parameters
        ----------
        x, y : dict
            Subgraph samples metric dictionaries for all levels.

        Returns
        -------
        kernel : number
            The kernel value.

        """
        return self.base_kernel_(x, y)

    def _calculate_MEC_(self, U):
        """Calculate the minimum enclosing cone for given U.

        Calculates the minimum a graph metric as the lovasz theta kernel or the
        svm-theta kernel, on a set of randomly sampled subgraphs, producing a
        dictionary of subgraph levels and sets.

        Parameters
        ----------
        U : np.array, ndim=2
            The adjacency matrix.

        Returns
        -------
        level_values : np.array, shape=(num_of_levels, 1)
            Returns for all levels the mean of lovasz_numbers for
            sampled subsets.

        """
        # Calculate subsets
        n = U.shape[1]
        samples_on_subsets = distribute_samples(n, self.subsets_size_range, self.n_samples)

        # Calculate level dictionary with lovasz values
        phi = np.zeros(shape=(self.subsets_size_range[1] - self.subsets_size_range[0]+1, 1))
        for (i, level) in enumerate(range(self.subsets_size_range[0], self.subsets_size_range[1] + 1)):
            v = samples_on_subsets.get(level, None)
            if v is not None:
                level_values = list()
                for k in range(v):
                    if level <= n:
                        indexes = self.random_state_.choice(n, level, replace=False)
                    else:
                        indexes = range(n)
                    # calculate the metrix value for that level
                    level_values.append(_minimum_cone_(U[:, indexes], self.random_state_))
                phi[i] = np.mean(level_values)

        return phi


def _calculate_lovasz_embeddings_(A):
    """Calculate the lovasz embeddings for the given graph.

    Parameters
    ----------
    A : np.array, ndim=2
        The adjacency matrix.

    Returns
    -------
    lovasz_theta : float
        Returns the lovasz theta number.

    optimization_solution : np.array
        Returns the slack variable solution
        of the primal optimization program.

    """
    if A.shape[0] == 1:
        return 1.0
    else:
        tf_adjm = (np.abs(A) <= min_weight)
        np.fill_diagonal(tf_adjm, False)
        i_list, j_list = np.nonzero(np.triu(tf_adjm.astype(int), k=1))

        nv = A.shape[0]
        ne = len(i_list)

        x_list = list()
        e_list = list()
        for (e, (i, j)) in enumerate(zip(i_list, j_list)):
            e_list.append(int(e)), x_list.append(int(i*nv+j))
            if tf_adjm[i, j]:
                e_list.append(int(e)), x_list.append(int(i*nv+j))

    # Add on the last row, diagonal elements
    e_list = e_list+(nv*[ne])
    x_list = x_list+[int(i*nv + i) for i in range(nv)]

    # initialise g sparse (to values -1, based on two list that
    # define index and one that defines shape
    g_sparse = spmatrix(-1, x_list, e_list, (nv*nv, ne+1))

    # Initialise optimization parameters
    h = matrix(-1.0, (nv, nv))
    c = matrix([0.0]*ne + [1.0])

    # Solve the convex optimization problem
    sol = sdp(c, Gs=[g_sparse], hs=[h])

    return np.array(sol['ss'][0]), sol['x'][ne]


def _calculate_lovasz_labelling_(X, t, d):
    """Calculate the lovasz labelling for the extracted embeddings.

    Parameters
    ----------
    X : np.array, ndim=2
        The optimization solution of the sdp program.

    t : float
        Returns the lovasz theta number.

    d : int
        The number of features size.

    Returns
    -------
    lovasz_theta : float
        Returns the lovasz theta number.

    optimization_solution : np.array
        Returns the slack variable solution
        of the primal optimization program.

    """
    n = X.shape[0]

    try:
        V = cholesky(X)
    except LinAlgError:
        x = X.diagonal()
        x.setflags(write=True)
        x += 2*abs(eigvalsh(X, lower=False, eigvals=(0, 0))[0])
        V = cholesky(X)

    V = pad(V, [(0, d-n), (0, 0)], mode='constant', constant_values=0)

    c = np.zeros(shape=(d,))
    c[-1] = 1

    C = np.outer(c, np.ones(shape=(n,)))

    U = 1/sqrt(t)*(C+V)
    return U


def _minimum_cone_(U, rs):
    """Calculate the minimum cone.

    Computes the angle and center of the minimum cone, with its point
    in the origin, enclosing the vectors in A.

    Parameters
    ----------
    U : np.array, ndim=2
        The vectors that will be enclosed.

    rs : RandomState
        A RandomState object handling all randomness of the class.

    Returns
    -------
    angle_cosine : float
        Returns the cosine of the minimum angle.

    """
    n = U.shape[1]

    P = rs.permutation(n) - 1
    R = np.array([], dtype=int)
    c, _ = _b_minidisk_(U, P, R, rs)

    with np.errstate(divide='ignore'):
        c /= norm(c, 2)

    t = min(np.dot(U.T, c))
    if t > 1. and t < angle_precision:
        t = 1.

    elif t < -1. and t > -angle_precision:
        t = -1.

    return t


def _b_minidisk_(A, P, R, rs):
    """Calculate the minidisk.

    Implements Welzl's algorithm (*move-to-front*)

    Parameters
    ----------
    A : np.array, ndim=2
        The vectors that will be enclosed.

    P : np.array, ndim=1
        A random permutation of indeces.

    R : np.array, ndim=1
        A subset of vectors that are enclosed.

    rs : RandomState
        A RandomState object handling all randomness of the class.

    Returns
    -------
    c : np.array, ndim=1
        The center vector C of the cone as it is being minimized.

    r : int
        The cone radius as it is beeing minimized.

    """
    d, nP, nR = A.shape[0], P.shape[0], R.shape[0]

    # original algorithm
    if nP == 0 or nR == d+1:
        if nR == 0:
            c, r = np.zeros(shape=(d,)), 0
        else:
            c, r = _fitball_(A[:, R])
    else:
        p = P[rs.randint(0, nP)]
        P_prime = np.delete(P, np.where(P == p))
        c, r = _b_minidisk_(A, P_prime, R, rs)
        if norm(A[:, p] - c, 2) - r > tolerance:
            # if not inside ball
            if p not in R:
                R_prime = pad(R, [(0, 1)], mode='constant', constant_values=p)
                c, r = _b_minidisk_(A, P_prime, R_prime, rs)
    return c, r


def _fitball_(A):
    """Fit the minimum ball.

    Parameters
    ----------
    A : np.array, ndim=2
        The vectors that will be enclosed inside the ball.

    Returns
    -------
    c : np.array, ndim=1
        The center vector C of the ball.

    r : int
        The ball radius.

    """
    d = A.shape[0]
    n = A.shape[1]

    if n == 1:
        c, r = A[:, 0], 0
    else:
        Q = A-np.outer(A[:, 0], np.ones(shape=(n, 1)))
        B = 2*np.dot(Q.T, Q)
        b = B.diagonal()/2

        L = solve(B[1:, :][:, 1:], b[1:])
        L = pad(L, [(1, 0)], mode='constant', constant_values=0)

        C = np.zeros(shape=(d,))

        for i in range(1, n):
            C = C + L[i]*Q[:, i]

        r = np.sqrt(np.dot(C, C))
        c = C + A[:, 1]

    return c, r


def inner_product(x, y):
    return x.T.dot(y)
